@@ -154,22 +154,30 @@ static int do_spawn(unsigned logopt, unsigned int wait,
 
 	f = fork();
 	if (f == 0) {
+		char **pargv = (char **) argv;
+		int loc = 0;
+
 		reset_signals();
 		close(pipefd[0]);
 		dup2(pipefd[1], STDOUT_FILENO);
 		dup2(pipefd[1], STDERR_FILENO);
 		close(pipefd[1]);
 
-		/* Bind mount - check target exists */
-		if (use_access) {
-			char **pargv = (char **) argv;
-			int argc = 0;
-			pid_t pgrp = getpgrp();
+		/* what to mount must always be second last */
+		while (*pargv++)
+			loc++;
+		loc -= 2;
 
-			/* what to mount must always be second last */
-			while (*pargv++)
-				argc++;
-			argc -= 2;
+		/*
+		 * If the mount location starts with a "/" then it is
+		 * a local path. In this case it is a bind mount, a
+		 * loopback mount or a file system that uses a local
+		 * path so we need to check for dependent mounts.
+		 *
+		 * I hope host names are never allowed "/" as first char
+		 */
+		if (use_access && *(argv[loc]) == '/') {
+			pid_t pgrp = getpgrp();
 
 			/*
 			 * Pretend to be requesting user and set non-autofs
@@ -181,13 +189,20 @@ static int do_spawn(unsigned logopt, unsigned int wait,
 			}
 			setpgrp();
 
-			/* Trigger the recursive mount */
-			if (access(argv[argc], F_OK) == -1)
-				_exit(errno);
+			/*
+			 * Trigger the recursive mount.
+			 *
+			 * Ignore the access(2) return code as there may be
+			 * multiple waiters for this mount and we need to
+			 * let the  VFS handle access returns to each
+			 * individual waiter.
+			 */
+			access(argv[loc], F_OK);
 
 			seteuid(0);
 			setegid(0);
-			setpgid(0, pgrp);
+			if (pgrp >= 0)
+				setpgid(0, pgrp);
 		}
 
 		execv(prog, (char *const *) argv);
@@ -305,13 +320,14 @@ int spawn_mount(unsigned logopt, ...)
 	unsigned int options;
 	unsigned int retries = MTAB_LOCK_RETRIES;
 	int update_mtab = 1, ret, printed = 0;
-	char buf[PATH_MAX];
+	unsigned int wait = defaults_get_mount_wait();
+	char buf[PATH_MAX + 1];
 
 	/* If we use mount locking we can't validate the location */
 #ifdef ENABLE_MOUNT_LOCKING
 	options = SPAWN_OPT_LOCK;
 #else
-	options = SPAWN_OPT_NONE;
+	options = SPAWN_OPT_ACCESS;
 #endif
 
 	va_start(arg, logopt);
@@ -321,7 +337,8 @@ int spawn_mount(unsigned logopt, ...)
 	ret = readlink(_PATH_MOUNTED, buf, PATH_MAX);
 	if (ret != -1) {
 		buf[ret] = '\0';
-		if (!strcmp(buf, _PROC_MOUNTS)) {
+		if (!strcmp(buf, _PROC_MOUNTS) ||
+		    !strcmp(buf, _PROC_SELF_MOUNTS)) {
 			debug(logopt,
 			      "mtab link detected, passing -n to mount");
 			argc++;
@@ -330,7 +347,7 @@ int spawn_mount(unsigned logopt, ...)
 	}
 
 	/* Alloc 1 extra slot in case we need to use the "-f" option */
-	if (!(argv = alloca(sizeof(char *) * argc + 2)))
+	if (!(argv = alloca(sizeof(char *) * (argc + 2))))
 		return -1;
 
 	argv[0] = arg0;
@@ -343,19 +360,24 @@ int spawn_mount(unsigned logopt, ...)
 		p = argv + 2;
 	}
 	while ((*p = va_arg(arg, char *))) {
-		if (options == SPAWN_OPT_NONE && !strcmp(*p, "-o")) {
+		if (options == SPAWN_OPT_ACCESS && !strcmp(*p, "-t")) {
 			*(++p) = va_arg(arg, char *);
 			if (!*p)
 				break;
-			if (strstr(*p, "loop"))
-				options = SPAWN_OPT_ACCESS;
+			/*
+			 * A cifs mount location begins with a "/" but
+			 * is not a local path, so don't try to resolve
+			 * it. Mmmm ... does anyone use smbfs these days?
+			 */
+			if (strstr(*p, "cifs"))
+				options = SPAWN_OPT_NONE;
 		}
 		p++;
 	}
 	va_end(arg);
 
 	while (retries--) {
-		ret = do_spawn(logopt, -1, options, prog, (const char **) argv);
+		ret = do_spawn(logopt, wait, options, prog, (const char **) argv);
 		if (ret & MTAB_NOTUPDATED) {
 			struct timespec tm = {3, 0};
 
@@ -427,7 +449,7 @@ int spawn_bind_mount(unsigned logopt, ...)
 	unsigned int options;
 	unsigned int retries = MTAB_LOCK_RETRIES;
 	int update_mtab = 1, ret, printed = 0;
-	char buf[PATH_MAX];
+	char buf[PATH_MAX + 1];
 
 	/* If we use mount locking we can't validate the location */
 #ifdef ENABLE_MOUNT_LOCKING
@@ -447,7 +469,8 @@ int spawn_bind_mount(unsigned logopt, ...)
 	ret = readlink(_PATH_MOUNTED, buf, PATH_MAX);
 	if (ret != -1) {
 		buf[ret] = '\0';
-		if (!strcmp(buf, _PROC_MOUNTS)) {
+		if (!strcmp(buf, _PROC_MOUNTS) ||
+		    !strcmp(buf, _PROC_SELF_MOUNTS)) {
 			debug(logopt,
 			      "mtab link detected, passing -n to mount");
 			argc++;
@@ -455,7 +478,7 @@ int spawn_bind_mount(unsigned logopt, ...)
 		}
 	}
 
-	if (!(argv = alloca(sizeof(char *) * argc + 2)))
+	if (!(argv = alloca(sizeof(char *) * (argc + 2))))
 		return -1;
 
 	argv[0] = arg0;
@@ -534,7 +557,7 @@ int spawn_umount(unsigned logopt, ...)
 	unsigned int retries = MTAB_LOCK_RETRIES;
 	int update_mtab = 1, ret, printed = 0;
 	unsigned int wait = defaults_get_umount_wait();
-	char buf[PATH_MAX];
+	char buf[PATH_MAX + 1];
 
 #ifdef ENABLE_MOUNT_LOCKING
 	options = SPAWN_OPT_LOCK;
@@ -549,7 +572,8 @@ int spawn_umount(unsigned logopt, ...)
 	ret = readlink(_PATH_MOUNTED, buf, PATH_MAX);
 	if (ret != -1) {
 		buf[ret] = '\0';
-		if (!strcmp(buf, _PROC_MOUNTS)) {
+		if (!strcmp(buf, _PROC_MOUNTS) ||
+		    !strcmp(buf, _PROC_SELF_MOUNTS)) {
 			debug(logopt,
 			      "mtab link detected, passing -n to mount");
 			argc++;
