@@ -84,7 +84,8 @@ int load_autofs4_module(void);
 #define MTAB_NOTUPDATED 0x1000			/* mtab succeded but not updated */
 #define NOT_MOUNTED     0x0100			/* path notmounted */
 #define MNT_FORCE_FAIL	-1
-#define _PROC_MOUNTS	"/proc/mounts"
+#define _PROC_MOUNTS		"/proc/mounts"
+#define _PROC_SELF_MOUNTS	"/proc/self/mounts"
 
 /* Constants for lookup modules */
 
@@ -114,6 +115,8 @@ int load_autofs4_module(void);
 #define DB(x)           do { } while(0)
 #endif
 
+#define min(a, b)	(a <= b ? a : b)
+
 /* Forward declaraion */
 struct autofs_point; 
 
@@ -185,10 +188,12 @@ struct mapent *cache_lookup_distinct(struct mapent_cache *mc, const char *key);
 struct mapent *cache_lookup_offset(const char *prefix, const char *offset, int start, struct list_head *head);
 struct mapent *cache_partial_match(struct mapent_cache *mc, const char *prefix);
 int cache_add(struct mapent_cache *mc, struct map_source *ms, const char *key, const char *mapent, time_t age);
-int cache_add_offset(struct mapent_cache *mc, const char *mkey, const char *key, const char *mapent, time_t age);
+int cache_update_offset(struct mapent_cache *mc, const char *mkey, const char *key, const char *mapent, time_t age);
+void cache_update_negative(struct mapent_cache *mc, struct map_source *ms, const char *key, time_t timeout);
 int cache_set_parents(struct mapent *mm);
 int cache_update(struct mapent_cache *mc, struct map_source *ms, const char *key, const char *mapent, time_t age);
 int cache_delete(struct mapent_cache *mc, const char *key);
+int cache_delete_offset(struct mapent_cache *mc, const char *key);
 void cache_multi_readlock(struct mapent *me);
 void cache_multi_writelock(struct mapent *me);
 void cache_multi_unlock(struct mapent *me);
@@ -229,7 +234,7 @@ int rmdir_path(struct autofs_point *ap, const char *path, dev_t dev);
 #define AUTOFS_LOOKUP_VERSION 5
 
 #define KEY_MAX_LEN    NAME_MAX
-#define MAPENT_MAX_LEN 4095
+#define MAPENT_MAX_LEN 16384
 #define PARSE_MAX_BUF	KEY_MAX_LEN + MAPENT_MAX_LEN + 2
 
 int lookup_nss_read_master(struct master *master, time_t age);
@@ -333,7 +338,9 @@ int ncat_path(char *buf, size_t len,
 
 /* Core automount definitions */
 
+#ifndef MNT_DETACH
 #define MNT_DETACH	0x00000002	/* Just detach from the tree */
+#endif
 
 struct startup_cond {
 	pthread_mutex_t mutex;
@@ -449,6 +456,9 @@ struct kernel_mod_version {
 /* Don't use bind mounts even when system supports them */
 #define MOUNT_FLAG_NOBIND		0x0020
 
+/* Use symlinks instead of bind mounting local mounts */
+#define MOUNT_FLAG_SYMLINK		0x0040
+
 struct autofs_point {
 	pthread_t thid;
 	char *path;			/* Mount point name */
@@ -459,7 +469,6 @@ struct autofs_point {
 	dev_t dev;			/* "Device" number assigned by kernel */
 	struct master_mapent *entry;	/* Master map entry for this mount */
 	unsigned int type;		/* Type of map direct or indirect */
-	time_t exp_timeout;		/* Timeout for expiring mounts */
 	time_t exp_runfreq;		/* Frequency for polling for timeouts */
 	time_t negative_timeout;	/* timeout in secs for failed mounts */
 	unsigned int flags;		/* autofs mount flags */
@@ -493,7 +502,7 @@ void *expire_proc_indirect(void *);
 void *expire_proc_direct(void *);
 int expire_offsets_direct(struct autofs_point *ap, struct mapent *me, int now);
 int mount_autofs_indirect(struct autofs_point *ap, const char *root);
-int do_mount_autofs_direct(struct autofs_point *ap, struct mnt_list *mnts, struct mapent *me);
+int do_mount_autofs_direct(struct autofs_point *ap, struct mnt_list *mnts, struct mapent *me, time_t timeout);
 int mount_autofs_direct(struct autofs_point *ap);
 int mount_autofs_offset(struct autofs_point *ap, struct mapent *me, const char *root, const char *offset);
 void submount_signal_parent(struct autofs_point *ap, unsigned int success);
@@ -539,7 +548,8 @@ static inline void check_cloexec(int fd)
 {
 	if (cloexec_works == 0) {
 		int fl = fcntl(fd, F_GETFD);
-		cloexec_works = (fl & FD_CLOEXEC) ? 1 : -1;
+		if (fl != -1)
+			cloexec_works = (fl & FD_CLOEXEC) ? 1 : -1;
 	}
 	if (cloexec_works > 0)
 		return;
