@@ -27,13 +27,30 @@
 #include <net/if.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
-#include <rpcsvc/ypclnt.h>
 #include <errno.h>
 #include <sys/ioctl.h>
 #include <ctype.h>
 #include <pthread.h>
 #include <poll.h>
 
+#ifdef WITH_LIBTIRPC
+#undef auth_destroy
+#define auth_destroy(auth)                                              \
+                do {                                                    \
+                        int refs;                                       \
+                        if ((refs = auth_put((auth))) == 0)             \
+                                ((*((auth)->ah_ops->ah_destroy))(auth));\
+                } while (0)
+#endif
+
+#ifdef WITH_LIBTIRPC
+const rpcprog_t rpcb_prog = RPCBPROG;
+const rpcvers_t rpcb_version = RPCBVERS;
+#else
+const rpcprog_t rpcb_prog = PMAPPROG;
+const rpcvers_t rpcb_version = PMAPVERS;
+#endif
+
 #include "mount.h"
 #include "rpc_subs.h"
 #include "automount.h"
@@ -53,80 +70,15 @@
 /* Get numeric value of the n bits starting at position p */
 #define getbits(x, p, n)      ((x >> (p + 1 - n)) & ~(~0 << n))
 
+static const rpcvers_t mount_vers[] = {
+        MOUNTVERS_NFSV3,
+        MOUNTVERS_POSIX,
+        MOUNTVERS,
+};
+
+static int connect_nb(int, struct sockaddr *, socklen_t, struct timeval *);
 inline void dump_core(void);
 
-static CLIENT *rpc_clntudp_create(struct sockaddr *addr, struct conn_info *info, int *fd)
-{
-	struct sockaddr_in *in4_raddr;
-	struct sockaddr_in6 *in6_raddr;
-	CLIENT *client = NULL;
-
-	switch (addr->sa_family) {
-	case AF_INET:
-		in4_raddr = (struct sockaddr_in *) addr;
-		in4_raddr->sin_port = htons(info->port);
-		client = clntudp_bufcreate(in4_raddr,
-					   info->program, info->version,
-					   info->timeout, fd,
-					   info->send_sz, info->recv_sz);
-		break;
-
-	case AF_INET6:
-#ifndef INET6
-		/* Quiet compile warning */
-		in6_raddr = NULL;
-#else
-		in6_raddr = (struct sockaddr_in6 *) addr;
-		in6_raddr->sin6_port = htons(info->port);
-		client = clntudp6_bufcreate(in6_raddr,
-					    info->program, info->version,
-					    info->timeout, fd,
-					    info->send_sz, info->recv_sz);
-#endif
-		break;
-
-	default:
-		break;
-	}
-
-	return client;
-}
-
-static CLIENT *rpc_clnttcp_create(struct sockaddr *addr, struct conn_info *info, int *fd)
-{
-	struct sockaddr_in *in4_raddr;
-	struct sockaddr_in6 *in6_raddr;
-	CLIENT *client = NULL;
-
-	switch (addr->sa_family) {
-	case AF_INET:
-		in4_raddr = (struct sockaddr_in *) addr;
-		in4_raddr->sin_port = htons(info->port);
-		client = clnttcp_create(in4_raddr,
-					info->program, info->version, fd,
-					info->send_sz, info->recv_sz);
-		break;
-
-	case AF_INET6:
-#ifndef INET6
-		/* Quiet compile warning */
-		in6_raddr = NULL;
-#else
-		in6_raddr = (struct sockaddr_in6 *) addr;
-		in6_raddr->sin6_port = htons(info->port);
-		client = clnttcp6_create(in6_raddr,
-					 info->program, info->version, fd,
-					 info->send_sz, info->recv_sz);
-#endif
-		break;
-
-	default:
-		break;
-	}
-
-	return client;
-}
-
 /*
  *  Perform a non-blocking connect on the socket fd.
  *
@@ -141,11 +93,11 @@ static int connect_nb(int fd, struct sockaddr *addr, socklen_t len, struct timev
 
 	flags = fcntl(fd, F_GETFL, 0);
 	if (flags < 0)
-		return -1;
+		return -errno;
 
 	ret = fcntl(fd, F_SETFL, flags | O_NONBLOCK);
 	if (ret < 0)
-		return -1;
+		return -errno;
 
 	/* 
 	 * From here on subsequent sys calls could change errno so
@@ -161,6 +113,13 @@ static int connect_nb(int fd, struct sockaddr *addr, socklen_t len, struct timev
 	if (ret == 0)
 		goto done;
 
+	if (timeout != -1) {
+		if (timeout >= (INT_MAX - 1)/1000)
+			timeout = INT_MAX - 1;
+		else
+			timeout = timeout * 1000;
+	}
+
 	pfd[0].fd = fd;
 	pfd[0].events = POLLOUT;
 
@@ -207,16 +166,18 @@ done:
 	return ret;
 }
 
-static CLIENT *rpc_do_create_client(struct sockaddr *addr, struct conn_info *info, int *fd)
+#ifndef WITH_LIBTIRPC
+static int rpc_do_create_client(struct sockaddr *addr, struct conn_info *info, int *fd, CLIENT **client)
 {
-	CLIENT *client = NULL;
-	struct sockaddr *laddr;
+	CLIENT *clnt = NULL;
 	struct sockaddr_in in4_laddr;
-	struct sockaddr_in6 in6_laddr;
-	int type, proto;
+	struct sockaddr_in *in4_raddr;
+	int type, proto, ret;
 	socklen_t slen;
 
-	proto = info->proto->p_proto;
+	*client = NULL;
+
+	proto = info->proto;
 	if (proto == IPPROTO_UDP)
 		type = SOCK_DGRAM;
 	else
@@ -227,82 +188,212 @@ static CLIENT *rpc_do_create_client(struct sockaddr *addr, struct conn_info *inf
 	 * layer, it would bind to a reserved port, which has been shown
 	 * to exhaust the reserved port range in some situations.
 	 */
-	switch (addr->sa_family) {
-	case AF_INET:
-		in4_laddr.sin_family = AF_INET;
-		in4_laddr.sin_port = htons(0);
-		in4_laddr.sin_addr.s_addr = htonl(INADDR_ANY);
-		slen = sizeof(struct sockaddr_in);
-		laddr = (struct sockaddr *) &in4_laddr;
-		break;
+	in4_laddr.sin_family = AF_INET;
+	in4_laddr.sin_port = htons(0);
+	in4_laddr.sin_addr.s_addr = htonl(INADDR_ANY);
+	slen = sizeof(struct sockaddr_in);
 
-	case AF_INET6:
-#ifndef INET6
-		/* Quiet compiler */
-		in6_laddr.sin6_family = AF_INET6;
-		return NULL;
-#else
-		in6_laddr.sin6_family = AF_INET6;
-		in6_laddr.sin6_port = htons(0);
-		in6_laddr.sin6_addr = in6addr_any;
-		slen = sizeof(struct sockaddr_in6);
-		laddr = (struct sockaddr *) &in6_laddr;
-		break;
-#endif
-	default:
-		return NULL;
+	if (!info->client) {
+		struct sockaddr *laddr;
+
+		*fd = open_sock(addr->sa_family, type, proto);
+		if (*fd < 0)
+			return -errno;
+
+		laddr = (struct sockaddr *) &in4_laddr;
+		if (bind(*fd, laddr, slen) < 0)
+			return -errno;
 	}
 
-	switch (info->proto->p_proto) {
-	case IPPROTO_UDP:
-		if (!info->client) {
-			*fd = open_sock(addr->sa_family, type, proto);
-			if (*fd < 0)
-				return NULL;
+	in4_raddr = (struct sockaddr_in *) addr;
+	in4_raddr->sin_port = htons(info->port);
 
-			if (bind(*fd, laddr, slen) < 0) {
-				close(*fd);
-				return NULL;
-			}
-		}
-		client = rpc_clntudp_create(addr, info, fd);
+	switch (info->proto) {
+	case IPPROTO_UDP:
+		clnt = clntudp_bufcreate(in4_raddr,
+					 info->program, info->version,
+					 info->timeout, fd,
+					 info->send_sz, info->recv_sz);
 		break;
 
 	case IPPROTO_TCP:
-		if (!info->client) {
-			*fd = open_sock(addr->sa_family, type, proto);
-			if (*fd < 0)
-				return NULL;
+		ret = connect_nb(*fd, addr, slen, &info->timeout);
+		if (ret < 0)
+			return ret;
 
-			if (connect_nb(*fd, laddr, slen, &info->timeout) < 0) {
-				close(*fd);
-				return NULL;
-			}
-		}
-		client = rpc_clnttcp_create(addr, info, fd);
+		clnt = clnttcp_create(in4_raddr,
+				      info->program, info->version, fd,
+				      info->send_sz, info->recv_sz);
 		break;
 
 	default:
 		break;
 	}
 
-	return client;
+	*client = clnt;
+
+	return 0;
+}
+#else
+static int rpc_do_create_client(struct sockaddr *addr, struct conn_info *info, int *fd, CLIENT **client)
+{
+	CLIENT *clnt = NULL;
+	struct sockaddr_in in4_laddr;
+	struct sockaddr_in6 in6_laddr;
+	struct sockaddr *laddr = NULL;
+	struct netbuf nb_addr;
+	int type, proto;
+	size_t slen;
+	int ret;
+
+	*client = NULL;
+
+	proto = info->proto;
+	if (proto == IPPROTO_UDP)
+		type = SOCK_DGRAM;
+	else
+		type = SOCK_STREAM;
+
+	/*
+	 * bind to any unused port.  If we left this up to the rpc
+	 * layer, it would bind to a reserved port, which has been shown
+	 * to exhaust the reserved port range in some situations.
+	 */
+	if (addr->sa_family == AF_INET) {
+		struct sockaddr_in *in4_raddr = (struct sockaddr_in *) addr;
+		in4_laddr.sin_family = AF_INET;
+		in4_laddr.sin_port = htons(0);
+		in4_laddr.sin_addr.s_addr = htonl(INADDR_ANY);
+		laddr = (struct sockaddr *) &in4_laddr;
+		in4_raddr->sin_port = htons(info->port);
+		slen = sizeof(struct sockaddr_in);
+		/* Use rpcbind v2 for AF_INET */
+		if (info->program == rpcb_prog)
+			info->version = PMAPVERS;
+	} else if (addr->sa_family == AF_INET6) {
+		struct sockaddr_in6 *in6_raddr = (struct sockaddr_in6 *) addr;
+		in6_laddr.sin6_family = AF_INET6;
+		in6_laddr.sin6_port = htons(0);
+		in6_laddr.sin6_addr = in6addr_any;
+		laddr = (struct sockaddr *) &in6_laddr;
+		in6_raddr->sin6_port = htons(info->port);
+		slen = sizeof(struct sockaddr_in6);
+	} else
+		return -EINVAL;
+
+	/*
+	 * bind to any unused port.  If we left this up to the rpc layer,
+	 * it would bind to a reserved port, which has been shown to
+	 * exhaust the reserved port range in some situations.
+	 */
+	if (!info->client) {
+		*fd = open_sock(addr->sa_family, type, proto);
+		if (*fd < 0) {
+			ret = -errno;
+			return ret;
+		}
+
+		if (bind(*fd, laddr, slen) < 0) {
+			ret = -errno;
+			return ret;
+		}
+	}
+
+	nb_addr.maxlen = nb_addr.len = slen;
+	nb_addr.buf = addr;
+
+	if (info->proto == IPPROTO_UDP)
+		clnt = clnt_dg_create(*fd, &nb_addr,
+				      info->program, info->version,
+				      info->send_sz, info->recv_sz);
+	else if (info->proto == IPPROTO_TCP) {
+		ret = connect_nb(*fd, addr, slen, &info->timeout);
+		if (ret < 0)
+			return ret;
+		clnt = clnt_vc_create(*fd, &nb_addr,
+				      info->program, info->version,
+				      info->send_sz, info->recv_sz);
+	} else
+		return -EINVAL;
+
+	/* Our timeout is in seconds */
+	if (clnt && info->timeout.tv_sec)
+		clnt_control(clnt, CLSET_TIMEOUT, (void *) &info->timeout);
+
+	*client = clnt;
+
+	return 0;
+}
+#endif
+
+#if defined(HAVE_GETRPCBYNAME) || defined(HAVE_GETSERVBYNAME)
+static pthread_mutex_t rpcb_mutex = PTHREAD_MUTEX_INITIALIZER;
+#endif
+
+static rpcprog_t rpc_getrpcbyname(const rpcprog_t program)
+{
+#ifdef HAVE_GETRPCBYNAME
+	static const char *rpcb_pgmtbl[] = {
+		"rpcbind", "portmap", "portmapper", "sunrpc", NULL,
+	};
+	struct rpcent *entry;
+	rpcprog_t prog_number;
+	unsigned int i;
+
+	pthread_mutex_lock(&rpcb_mutex);
+	for (i = 0; rpcb_pgmtbl[i] != NULL; i++) {
+		entry = getrpcbyname(rpcb_pgmtbl[i]);
+		if (entry) {
+			prog_number = entry->r_number;
+			pthread_mutex_unlock(&rpcb_mutex);
+			return prog_number;
+		}
+	}
+	pthread_mutex_unlock(&rpcb_mutex);
+#endif
+	return program;
+}
+
+static unsigned short rpc_getrpcbport(const int proto)
+{
+#ifdef HAVE_GETSERVBYNAME
+	static const char *rpcb_netnametbl[] = {
+		"rpcbind", "portmapper", "sunrpc", NULL,
+	};
+	struct servent *entry;
+	struct protoent *p_ent;
+	unsigned short port;
+	unsigned int i;
+
+	pthread_mutex_lock(&rpcb_mutex);
+	p_ent = getprotobynumber(proto);
+	if (!p_ent)
+		goto done;
+	for (i = 0; rpcb_netnametbl[i] != NULL; i++) {
+		entry = getservbyname(rpcb_netnametbl[i], p_ent->p_name);
+		if (entry) {
+			port = entry->s_port;
+			pthread_mutex_unlock(&rpcb_mutex);
+			return port;
+		}
+	}
+done:
+	pthread_mutex_unlock(&rpcb_mutex);
+#endif
+	return (unsigned short) PMAPPORT;
 }
 
 /*
- * Create a UDP RPC client
+ * Create an RPC client
  */
-static CLIENT *create_udp_client(struct conn_info *info)
+static int create_client(struct conn_info *info, CLIENT **client)
 {
-	CLIENT *client = NULL;
 	struct addrinfo *ai, *haddr;
 	struct addrinfo hints;
 	int fd, ret;
 
-	if (info->proto->p_proto != IPPROTO_UDP)
-		return NULL;
-
 	fd = RPC_ANYSOCK;
+	*client = NULL;
 
 	if (info->client) {
 		if (!clnt_control(info->client, CLGET_FD, (char *) &fd)) {
@@ -316,11 +407,13 @@ static CLIENT *create_udp_client(struct conn_info *info)
 	}
 
 	if (info->addr) {
-		client = rpc_do_create_client(info->addr, info, &fd);
-		if (client)
+		ret = rpc_do_create_client(info->addr, info, &fd, client);
+		if (ret == 0)
 			goto done;
+		if (ret == -EHOSTUNREACH)
+			goto out_close;
 
-		if (!info->client) {
+		if (!info->client && fd != RPC_ANYSOCK) {
 			close(fd);
 			fd = RPC_ANYSOCK;
 		}
@@ -329,7 +422,10 @@ static CLIENT *create_udp_client(struct conn_info *info)
 	memset(&hints, 0, sizeof(hints));
 	hints.ai_flags = AI_ADDRCONFIG;
 	hints.ai_family = AF_UNSPEC;
-	hints.ai_socktype = SOCK_DGRAM;
+	if (info->proto == IPPROTO_UDP)
+		hints.ai_socktype = SOCK_DGRAM;
+	else
+		hints.ai_socktype = SOCK_STREAM;
 
 	ret = getaddrinfo(info->host, NULL, &hints, &ai);
 	if (ret) {
@@ -341,11 +437,18 @@ static CLIENT *create_udp_client(struct conn_info *info)
 
 	haddr = ai;
 	while (haddr) {
-		client = rpc_do_create_client(haddr->ai_addr, info, &fd);
-		if (client)
-			break;
+		if (haddr->ai_protocol != info->proto) {
+			haddr = haddr->ai_next;
+			continue;
+		}
 
-		if (!info->client) {
+		ret = rpc_do_create_client(haddr->ai_addr, info, &fd, client);
+		if (ret == 0)
+			break;
+		if (ret == -EHOSTUNREACH)
+			goto out_close;
+
+		if (!info->client && fd != RPC_ANYSOCK) {
 			close(fd);
 			fd = RPC_ANYSOCK;
 		}
@@ -355,38 +458,39 @@ static CLIENT *create_udp_client(struct conn_info *info)
 
 	freeaddrinfo(ai);
 
-	if (!client) {
-		info->client = NULL;
-		goto out_close;
-	}
 done:
-	/* Close socket fd on destroy, as is default for rpcowned fds */
-	if  (!clnt_control(client, CLSET_FD_CLOSE, NULL)) {
-		clnt_destroy(client);
+	if (!*client) {
 		info->client = NULL;
+		ret = -ENOTCONN;
 		goto out_close;
 	}
 
-	return client;
+	/* Close socket fd on destroy, as is default for rpcowned fds */
+	if  (!clnt_control(*client, CLSET_FD_CLOSE, NULL)) {
+		clnt_destroy(*client);
+		info->client = NULL;
+		ret = -ENOTCONN;
+		goto out_close;
+	}
+
+	return 0;
 
 out_close:
 	if (fd != -1)
 		close(fd);
-	return NULL;
+	return ret;
 }
 
 int rpc_udp_getclient(struct conn_info *info,
 		      unsigned int program, unsigned int version)
 {
-	struct protoent *pe_proto;
 	CLIENT *client;
+	int ret;
 
 	if (!info->client) {
-		pe_proto = getprotobyname("udp");
-		if (!pe_proto)
-			return 0;
-
-		info->proto = pe_proto;
+		info->proto = IPPROTO_UDP;
+		info->timeout.tv_sec = RPC_TOUT_UDP;
+		info->timeout.tv_usec = 0;
 		info->send_sz = UDPMSGSIZE;
 		info->recv_sz = UDPMSGSIZE;
 	}
@@ -394,14 +498,13 @@ int rpc_udp_getclient(struct conn_info *info,
 	info->program = program;
 	info->version = version;
 
-	client = create_udp_client(info);
-
-	if (!client)
-		return 0;
+	ret = create_client(info, &client);
+	if (ret < 0)
+		return ret;
 
 	info->client = client;
 
-	return 1;
+	return 0;
 }
 
 void rpc_destroy_udp_client(struct conn_info *info)
@@ -414,104 +517,16 @@ void rpc_destroy_udp_client(struct conn_info *info)
 	return;
 }
 
-/*
- * Create a TCP RPC client using non-blocking connect
- */
-static CLIENT *create_tcp_client(struct conn_info *info)
-{
-	CLIENT *client = NULL;
-	struct addrinfo *ai, *haddr;
-	struct addrinfo hints;
-	int fd, ret;
-
-	if (info->proto->p_proto != IPPROTO_TCP)
-		return NULL;
-
-	fd = RPC_ANYSOCK;
-
-	if (info->client) {
-		if (!clnt_control(info->client, CLGET_FD, (char *) &fd)) {
-			fd = RPC_ANYSOCK;
-			clnt_destroy(info->client);
-			info->client = NULL;
-		} else {
-			clnt_control(info->client, CLSET_FD_NCLOSE, NULL);
-			clnt_destroy(info->client);
-		}
-	}
-
-	if (info->addr) {
-		client = rpc_do_create_client(info->addr, info, &fd);
-		if (client)
-			goto done;
-
-		if (!info->client) {
-			close(fd);
-			fd = RPC_ANYSOCK;
-		}
-	}
-
-	memset(&hints, 0, sizeof(hints));
-	hints.ai_flags = AI_ADDRCONFIG;
-	hints.ai_family = AF_UNSPEC;
-	hints.ai_socktype = SOCK_STREAM;
-
-	ret = getaddrinfo(info->host, NULL, &hints, &ai);
-	if (ret) {
-		error(LOGOPT_ANY,
-		      "hostname lookup failed: %s", gai_strerror(ret));
-		info->client = NULL;
-		goto out_close;
-	}
-
-	haddr = ai;
-	while (haddr) {
-		client = rpc_do_create_client(haddr->ai_addr, info, &fd);
-		if (client)
-			break;
-
-		if (!info->client) {
-			close(fd);
-			fd = RPC_ANYSOCK;
-		}
-
-		haddr = haddr->ai_next;
-	}
-
-	freeaddrinfo(ai);
-
-	if (!client) {
-		info->client = NULL;
-		goto out_close;
-	}
-done:
-	/* Close socket fd on destroy, as is default for rpcowned fds */
-	if  (!clnt_control(client, CLSET_FD_CLOSE, NULL)) {
-		clnt_destroy(client);
-		info->client = NULL;
-		goto out_close;
-	}
-
-	return client;
-
-out_close:
-	if (fd != -1)
-		close(fd);
-	return NULL;
-}
-
 int rpc_tcp_getclient(struct conn_info *info,
 		      unsigned int program, unsigned int version)
 {
-	struct protoent *pe_proto;
 	CLIENT *client;
+	int ret;
 
 	if (!info->client) {
-		pe_proto = getprotobyname("tcp");
-		if (!pe_proto)
-			return 0;
-
-		info->proto = pe_proto;
+		info->proto = IPPROTO_TCP;
+		info->timeout.tv_sec = RPC_TOUT_TCP;
+		info->timeout.tv_usec = 0;
 		info->send_sz = 0;
 		info->recv_sz = 0;
 	}
@@ -519,14 +534,13 @@ int rpc_tcp_getclient(struct conn_info *info,
 	info->program = program;
 	info->version = version;
 
-	client = create_tcp_client(info);
-
-	if (!client)
-		return 0;
+	ret = create_client(info, &client);
+	if (ret < 0)
+		return ret;
 
 	info->client = client;
 
-	return 1;
+	return 0;
 }
 
 void rpc_destroy_tcp_client(struct conn_info *info)
@@ -556,22 +570,24 @@ void rpc_destroy_tcp_client(struct conn_info *info)
 
 int rpc_portmap_getclient(struct conn_info *info,
 			  const char *host, struct sockaddr *addr, size_t addr_len,
-			  const char *proto, unsigned int option)
+			  int proto, unsigned int option)
 {
-	struct protoent *pe_proto;
 	CLIENT *client;
-
-	pe_proto = getprotobyname(proto);
-	if (!pe_proto)
-		return 0;
+	int ret;
 
 	info->host = host;
 	info->addr = addr;
 	info->addr_len = addr_len;
-	info->program = PMAPPROG;
-	info->port = PMAPPORT;
-	info->version = PMAPVERS;
-	info->proto = pe_proto;
+	info->program = rpc_getrpcbyname(rpcb_prog);
+	info->port = ntohs(rpc_getrpcbport(proto));
+	/*
+	 * When using libtirpc we might need to change the rpcbind version
+	 * to qurey AF_INET addresses. Since we might not have an address
+	 * yet set AF_INET rpcbind version in rpc_do_create_client() when
+	 * we always have an address.
+	 */
+	info->version = rpcb_version;
+	info->proto = proto;
 	info->send_sz = RPCSMALLMSGSIZE;
 	info->recv_sz = RPCSMALLMSGSIZE;
 	info->timeout.tv_sec = PMAP_TOUT_UDP;
@@ -579,27 +595,26 @@ int rpc_portmap_getclient(struct conn_info *info,
 	info->close_option = option;
 	info->client = NULL;
 
-	if (pe_proto->p_proto == IPPROTO_TCP) {
+	if (info->proto == IPPROTO_TCP)
 		info->timeout.tv_sec = PMAP_TOUT_TCP;
-		client = create_tcp_client(info);
-	} else
-		client = create_udp_client(info);
 
-	if (!client)
-		return 0;
+	ret = create_client(info, &client);
+	if (ret < 0)
+		return ret;
 
 	info->client = client;
 
-	return 1;
+	return 0;
 }
 
-unsigned short rpc_portmap_getport(struct conn_info *info, struct pmap *parms)
+int rpc_portmap_getport(struct conn_info *info,
+			struct pmap *parms, unsigned short *port)
 {
 	struct conn_info pmap_info;
-	unsigned short port = 0;
 	CLIENT *client;
 	enum clnt_stat status;
-	int proto = info->proto->p_proto;
+	int proto = info->proto;
+	int ret;
 
 	memset(&pmap_info, 0, sizeof(struct conn_info));
 
@@ -614,20 +629,22 @@ unsigned short rpc_portmap_getport(struct conn_info *info, struct pmap *parms)
 		pmap_info.host = info->host;
 		pmap_info.addr = info->addr;
 		pmap_info.addr_len = info->addr_len;
-		pmap_info.port = PMAPPORT;
-		pmap_info.program = PMAPPROG;
-		pmap_info.version = PMAPVERS;
+		pmap_info.port = ntohs(rpc_getrpcbport(info->proto));
+		pmap_info.program = rpc_getrpcbyname(rpcb_prog);
+		/*
+		 * When using libtirpc we might need to change the rpcbind
+		 * version to qurey AF_INET addresses. Since we might not
+		 * have an address yet set AF_INET rpcbind version in
+		 * rpc_do_create_client() when we always have an address.
+		 */
+		pmap_info.version = rpcb_version;
 		pmap_info.proto = info->proto;
 		pmap_info.send_sz = RPCSMALLMSGSIZE;
 		pmap_info.recv_sz = RPCSMALLMSGSIZE;
 
-		if (proto == IPPROTO_TCP)
-			client = create_tcp_client(&pmap_info);
-		else
-			client = create_udp_client(&pmap_info);
-
-		if (!client)
-			return 0;
+		ret = create_client(&pmap_info, &client);
+		if (ret < 0)
+			return ret;
 	}
 
 	/*
@@ -641,7 +658,7 @@ unsigned short rpc_portmap_getport(struct conn_info *info, struct pmap *parms)
 	if (status == RPC_SUCCESS) {
 		status = clnt_call(client, PMAPPROC_GETPORT,
 				 (xdrproc_t) xdr_pmap, (caddr_t) parms,
-				 (xdrproc_t) xdr_u_short, (caddr_t) &port,
+				 (xdrproc_t) xdr_u_short, (caddr_t) port,
 				 pmap_info.timeout);
 	}
 
@@ -668,30 +685,31 @@ unsigned short rpc_portmap_getport(struct conn_info *info, struct pmap *parms)
 		clnt_destroy(client);
 	}
 
-	if (status != RPC_SUCCESS)
-		return 0;
+	if (status == RPC_TIMEDOUT)
+		return -ETIMEDOUT;
+	else if (status != RPC_SUCCESS)
+		return -EIO;
 
-	return port;
+	return 0;
 }
 
 int rpc_ping_proto(struct conn_info *info)
 {
 	CLIENT *client;
 	enum clnt_stat status;
-	int proto = info->proto->p_proto;
+	int proto = info->proto;
+	int ret;
 
 	if (info->client)
 		client = info->client;
 	else {
-		if (info->proto->p_proto == IPPROTO_UDP) {
+		if (info->proto == IPPROTO_UDP) {
 			info->send_sz = UDPMSGSIZE;
 			info->recv_sz = UDPMSGSIZE;
-			client = create_udp_client(info);
-		} else
-			client = create_tcp_client(info);
-
-		if (!client)
-			return 0;
+		}
+		ret = create_client(info, &client);
+		if (ret < 0)
+			return ret;
 	}
 
 	clnt_control(client, CLSET_TIMEOUT, (char *) &info->timeout);
@@ -724,22 +742,23 @@ int rpc_ping_proto(struct conn_info *info)
 		clnt_destroy(client);
 	}
 
-	if (status != RPC_SUCCESS)
-		return 0;
+	if (status == RPC_TIMEDOUT)
+		return -ETIMEDOUT;
+	else if (status != RPC_SUCCESS)
+		return -EIO;
 
 	return 1;
 }
 
-static unsigned int __rpc_ping(const char *host,
-				unsigned long version,
-				char *proto,
-				long seconds, long micros,
-				unsigned int option)
+static int __rpc_ping(const char *host,
+		      unsigned long version, int proto,
+		      long seconds, long micros, unsigned int option)
 {
-	unsigned int status;
+	int status;
 	struct conn_info info;
 	struct pmap parms;
 
+	info.proto = proto;
 	info.host = host;
 	info.addr = NULL;
 	info.addr_len = 0;
@@ -754,17 +773,13 @@ static unsigned int __rpc_ping(const char *host,
 
 	status = RPC_PING_FAIL;
 
-	info.proto = getprotobyname(proto);
-	if (!info.proto)
-		return status;
-
 	parms.pm_prog = NFS_PROGRAM;
 	parms.pm_vers = version;
-	parms.pm_prot = info.proto->p_proto;
+	parms.pm_prot = info.proto;
 	parms.pm_port = 0;
 
-	info.port = rpc_portmap_getport(&info, &parms);
-	if (!info.port)
+	status = rpc_portmap_getport(&info, &parms, &info.port);
+	if (status < 0)
 		return status;
 
 	status = rpc_ping_proto(&info);
@@ -776,22 +791,22 @@ int rpc_ping(const char *host, long seconds, long micros, unsigned int option)
 {
 	unsigned long vers3 = NFS3_VERSION;
 	unsigned long vers2 = NFS2_VERSION;
-	unsigned int status;
+	int status;
 
-	status = __rpc_ping(host, vers2, "udp", seconds, micros, option);
-	if (status)
+	status = __rpc_ping(host, vers2, IPPROTO_UDP, seconds, micros, option);
+	if (status > 0)
 		return RPC_PING_V2 | RPC_PING_UDP;
 
-	status = __rpc_ping(host, vers3, "udp", seconds, micros, option);
-	if (status)
+	status = __rpc_ping(host, vers3, IPPROTO_UDP, seconds, micros, option);
+	if (status > 0)
 		return RPC_PING_V3 | RPC_PING_UDP;
 
-	status = __rpc_ping(host, vers2, "tcp", seconds, micros, option);
-	if (status)
+	status = __rpc_ping(host, vers2, IPPROTO_TCP, seconds, micros, option);
+	if (status > 0)
 		return RPC_PING_V2 | RPC_PING_TCP;
 
-	status = __rpc_ping(host, vers3, "tcp", seconds, micros, option);
-	if (status)
+	status = __rpc_ping(host, vers3, IPPROTO_TCP, seconds, micros, option);
+	if (status > 0)
 		return RPC_PING_V3 | RPC_PING_TCP;
 
 	return status;
@@ -813,16 +828,15 @@ int rpc_time(const char *host,
 	double taken;
 	struct timeval start, end;
 	struct timezone tz;
-	char *proto = (ping_proto & RPC_PING_UDP) ? "udp" : "tcp";
+	int proto = (ping_proto & RPC_PING_UDP) ? IPPROTO_UDP : IPPROTO_TCP;
 	unsigned long vers = ping_vers;
 
 	gettimeofday(&start, &tz);
 	status = __rpc_ping(host, vers, proto, seconds, micros, option);
 	gettimeofday(&end, &tz);
 
-	if (!status) {
-		return 0;
-	}
+	if (status == RPC_PING_FAIL || status < 0)
+		return status;
 
 	taken = elapsed(start, end);
 
@@ -836,17 +850,17 @@ static int rpc_get_exports_proto(struct conn_info *info, exports *exp)
 {
 	CLIENT *client;
 	enum clnt_stat status;
-	int proto = info->proto->p_proto;
+	int proto = info->proto;
 	unsigned int option = info->close_option;
+	int vers_entry;
+	int ret;
 
-	if (info->proto->p_proto == IPPROTO_UDP) {
+	if (info->proto == IPPROTO_UDP) {
 		info->send_sz = UDPMSGSIZE;
 		info->recv_sz = UDPMSGSIZE;
-		client = create_udp_client(info);
-	} else
-		client = create_tcp_client(info);
-
-	if (!client)
+	}
+	ret = create_client(info, &client);
+	if (ret < 0)
 		return 0;
 
 	clnt_control(client, CLSET_TIMEOUT, (char *) &info->timeout);
@@ -854,10 +868,19 @@ static int rpc_get_exports_proto(struct conn_info *info, exports *exp)
 
 	client->cl_auth = authunix_create_default();
 
-	status = clnt_call(client, MOUNTPROC_EXPORT,
-			 (xdrproc_t) xdr_void, NULL,
-			 (xdrproc_t) xdr_exports, (caddr_t) exp,
-			 info->timeout);
+	vers_entry = 0;
+	while (1) {
+		status = clnt_call(client, MOUNTPROC_EXPORT,
+				 (xdrproc_t) xdr_void, NULL,
+				 (xdrproc_t) xdr_exports, (caddr_t) exp,
+				 info->timeout);
+		if (status == RPC_SUCCESS)
+			break;
+		if (++vers_entry > 2)
+			break;
+		CLNT_CONTROL(client, CLSET_VERS,
+			    (void *) &mount_vers[vers_entry]);
+	}
 
 	/* Only play with the close options if we think it completed OK */
 	if (proto == IPPROTO_TCP && status == RPC_SUCCESS) {
@@ -926,7 +949,7 @@ exports rpc_get_exports(const char *host, long seconds, long micros, unsigned in
 	info.addr = NULL;
 	info.addr_len = 0;
 	info.program = MOUNTPROG;
-	info.version = MOUNTVERS;
+	info.version = mount_vers[0];
 	info.send_sz = 0;
 	info.recv_sz = 0;
 	info.timeout.tv_sec = seconds;
@@ -939,14 +962,12 @@ exports rpc_get_exports(const char *host, long seconds, long micros, unsigned in
 	parms.pm_port = 0;
 
 	/* Try UDP first */
-	info.proto = getprotobyname("udp");
-	if (!info.proto)
-		goto try_tcp;
+	info.proto = IPPROTO_UDP;
 
-	parms.pm_prot = info.proto->p_proto;
+	parms.pm_prot = info.proto;
 
-	info.port = rpc_portmap_getport(&info, &parms);
-	if (!info.port)
+	status = rpc_portmap_getport(&info, &parms, &info.port);
+	if (status < 0)
 		goto try_tcp;
 
 	memset(&exportlist, '\0', sizeof(exportlist));
@@ -956,14 +977,12 @@ exports rpc_get_exports(const char *host, long seconds, long micros, unsigned in
 		return exportlist;
 
 try_tcp:
-	info.proto = getprotobyname("tcp");
-	if (!info.proto)
-		return NULL;
+	info.proto = IPPROTO_TCP;
 
-	parms.pm_prot = info.proto->p_proto;
+	parms.pm_prot = info.proto;
 
-	info.port = rpc_portmap_getport(&info, &parms);
-	if (!info.port)
+	status = rpc_portmap_getport(&info, &parms, &info.port);
+	if (status < 0)
 		return NULL;
 
 	memset(&exportlist, '\0', sizeof(exportlist));
@@ -975,6 +994,26 @@ try_tcp:
 	return exportlist;
 }
 
+const char *get_addr_string(struct sockaddr *sa, char *name, socklen_t len)
+{
+	void *addr;
+
+	if (len < INET6_ADDRSTRLEN)
+		return NULL;
+
+	if (sa->sa_family == AF_INET) {
+		struct sockaddr_in *ipv4 = (struct sockaddr_in *) sa;
+		addr = &(ipv4->sin_addr);
+	} else if (sa->sa_family == AF_INET6) {
+		struct sockaddr_in6 *ipv6 = (struct sockaddr_in6 *) sa;
+		addr = &(ipv6->sin6_addr);
+	} else {
+		return NULL;
+	}
+
+	return inet_ntop(sa->sa_family, addr, name, len);
+}
+
 #if 0
 #include <stdio.h>
 
