@@ -18,8 +18,6 @@
 #include <stdio.h>
 #include <malloc.h>
 #include <netdb.h>
-#include <fcntl.h>
-#include <unistd.h>
 #include <stdlib.h>
 #include <string.h>
 #include <ctype.h>
@@ -31,6 +29,8 @@
 #include <sys/vfs.h>
 #include <sys/utsname.h>
 #include <netinet/in.h>
+#include <sys/mount.h>
+#include <linux/fs.h>
 
 #define MODULE_PARSE
 #include "automount.h"
@@ -41,6 +41,24 @@ int parse_version = AUTOFS_PARSE_VERSION;	/* Required by protocol */
 
 static struct mount_mod *mount_nfs = NULL;
 static int init_ctr = 0;
+static int macro_init_done = 0;
+static pthread_mutex_t instance_mutex = PTHREAD_MUTEX_INITIALIZER;
+
+static void instance_mutex_lock(void)
+{
+	int status = pthread_mutex_lock(&instance_mutex);
+	if (status)
+		fatal(status);
+}
+
+static void instance_mutex_unlock(void)
+{
+	int status = pthread_mutex_unlock(&instance_mutex);
+	if (status)
+		fatal(status);
+}
+
+extern const char *global_options;
 
 struct parse_context {
 	char *optstr;		/* Mount options */
@@ -65,6 +83,9 @@ static struct parse_context default_context = {
 	1			/* Do slashify_colons */
 };
 
+int destroy_logpri_fifo(struct autofs_point *ap);
+static char *concat_options(char *left, char *right);
+
 /* Free all storage associated with this context */
 static void kill_context(struct parse_context *ctxt)
 {
@@ -78,44 +99,6 @@ static void kill_context(struct parse_context *ctxt)
 	free(ctxt);
 }
 
-static struct substvar *addstdenv(struct substvar *sv)
-{
-	struct substvar *list = sv;
-	struct thread_stdenv_vars *tsv;
-	char numbuf[16];
-
-	tsv = pthread_getspecific(key_thread_stdenv_vars);
-	if (tsv) {
-		int ret;
-		long num;
-
-		num = (long) tsv->uid;
-		ret = sprintf(numbuf, "%ld", num);
-		if (ret > 0)
-			list = macro_addvar(list, "UID", 3, numbuf);
-		num = (long) tsv->gid;
-		ret = sprintf(numbuf, "%ld", num);
-		if (ret > 0)
-			list = macro_addvar(list, "GID", 3, numbuf);
-		list = macro_addvar(list, "USER", 4, tsv->user);
-		list = macro_addvar(list, "GROUP", 5, tsv->group);
-		list = macro_addvar(list, "HOME", 4, tsv->home);
-	}
-	return list;
-}
-
-static struct substvar *removestdenv(struct substvar *sv)
-{
-	struct substvar *list = sv;
-
-	list = macro_removevar(list, "UID", 3);
-	list = macro_removevar(list, "USER", 4);
-	list = macro_removevar(list, "HOME", 4);
-	list = macro_removevar(list, "GID", 3);
-	list = macro_removevar(list, "GROUP", 5);
-	return list;
-}
-
 /* 
  * $- and &-expand a Sun-style map entry and return the length of the entry.
  * If "dst" is NULL, just count the length.
@@ -136,30 +119,24 @@ int expandsunent(const char *src, char *dst, const char *key,
 		case '&':
 			l = strlen(key);
 			/*
-			 * In order to ensure that any spaces in the key
-			 * re preserved, we need to escape them here.
+			 * In order to ensure that any isspace() characters
+			 * in the key are preserved, we need to escape them
+			 * here.
 			 */
-			if (strchr(key, ' ')) {
-				char *keyp = key;
-				while (*keyp) {
-					if (isspace(*keyp)) {
-						if (dst) {
-							*dst++ = '\\';
-							*dst++ = *keyp++;
-						} else
-							keyp++;
-						l++;
-					} else {
-						if (dst)
-							*dst++ = *keyp++;
-						else
-							keyp++;
-					}
-				}
-			} else {
-				if (dst) {
-					strcpy(dst, key);
-					dst += l;
+			const char *keyp = key;
+			while (*keyp) {
+				if (isspace(*keyp)) {
+					if (dst) {
+						*dst++ = '\\';
+						*dst++ = *keyp++;
+					} else
+						keyp++;
+					l++;
+				} else {
+					if (dst)
+						*dst++ = *keyp++;
+					else
+						keyp++;
 				}
 			}
 			len += l;
@@ -182,8 +159,7 @@ int expandsunent(const char *src, char *dst, const char *key,
 						dst += l;
 					}
 					len += l;
-				} else
-					return 0;
+				}
 				src = p + 1;
 			} else {
 				p = src;
@@ -197,8 +173,7 @@ int expandsunent(const char *src, char *dst, const char *key,
 						dst += l;
 					}
 					len += l;
-				} else
-					return 0;
+				}
 				src = p;
 			}
 			break;
@@ -238,7 +213,9 @@ int expandsunent(const char *src, char *dst, const char *key,
 				*(dst++) = 
 				  (seen_colons && slashify_colons) ? '/' : ':';
 			len++;
-			seen_colons = 1;
+			/* Were looking for the colon preceeding a path */
+			if (*src == '/')
+				seen_colons = 1;
 			break;
 
 		default:
@@ -260,21 +237,25 @@ int parse_init(int argc, const char *const *argv, void **context)
 {
 	struct parse_context *ctxt;
 	char buf[MAX_ERR_BUF];
-	char *noptstr, *def, *val, *macros;
+	char *noptstr, *def, *val, *macros, *gbl_options;
 	const char *xopt;
 	int optlen, len, offset;
 	int i, bval;
+	unsigned int append_options;
 
 	/* Get processor information for predefined escapes */
-
-	if (!init_ctr)
+	macro_lock();
+	if (!macro_init_done) {
+		macro_init_done = 1;
 		macro_init();
+	}
+	macro_unlock();
 
 	/* Set up context and escape chain */
 
 	if (!(ctxt = (struct parse_context *) malloc(sizeof(struct parse_context)))) {
 		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-		crit(LOGOPT_ANY, MODPREFIX "malloc: %s", estr);
+		logerr(MODPREFIX "malloc: %s", estr);
 		*context = NULL;
 		return 1;
 	}
@@ -299,7 +280,7 @@ int parse_init(int argc, const char *const *argv, void **context)
 
 				if (!def) {
 					char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-					error(LOGOPT_ANY, MODPREFIX "strdup: %s", estr);
+					logerr(MODPREFIX "strdup: %s", estr);
 					break;
 				}
 
@@ -371,20 +352,22 @@ int parse_init(int argc, const char *const *argv, void **context)
 			if (ctxt->optstr) {
 				noptstr =
 				    (char *) realloc(ctxt->optstr, optlen + len + 2);
-				if (!noptstr)
-					break;
-				noptstr[optlen] = ',';
-				strcpy(noptstr + optlen + 1, argv[i] + offset);
-				optlen += len + 1;
+				if (noptstr) {
+					noptstr[optlen] = ',';
+					strcpy(noptstr + optlen + 1, argv[i] + offset);
+					optlen += len + 1;
+				}
 			} else {
 				noptstr = (char *) malloc(len + 1);
-				strcpy(noptstr, argv[i] + offset);
-				optlen = len;
+				if (noptstr) {
+					strcpy(noptstr, argv[i] + offset);
+					optlen = len;
+				}
 			}
 			if (!noptstr) {
 				char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
 				kill_context(ctxt);
-				crit(LOGOPT_ANY, MODPREFIX "%s", estr);
+				logerr(MODPREFIX "%s", estr);
 				*context = NULL;
 				return 1;
 			}
@@ -392,24 +375,51 @@ int parse_init(int argc, const char *const *argv, void **context)
 		}
 	}
 
+	gbl_options = NULL;
+	if (global_options) {
+		if (ctxt->optstr && strstr(ctxt->optstr, global_options))
+			goto options_done;
+		gbl_options = strdup(global_options);
+	}
+
+	if (gbl_options) {
+		append_options = defaults_get_append_options();
+		if (append_options) {
+			char *tmp = concat_options(gbl_options, ctxt->optstr);
+			if (!tmp) {
+				char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
+				logerr(MODPREFIX "concat_options: %s", estr);
+				free(gbl_options);
+			} else
+				ctxt->optstr = tmp;
+		} else {
+			if (!ctxt->optstr)
+				ctxt->optstr = gbl_options;
+			else
+				free(gbl_options);
+		}
+	}
+options_done:
 	debug(LOGOPT_NONE,
 	      MODPREFIX "init gathered global options: %s", ctxt->optstr);
 
 	/* We only need this once.  NFS mounts are so common that we cache
 	   this module. */
-	if (!mount_nfs) {
+	instance_mutex_lock();
+	if (mount_nfs)
+		init_ctr++;
+	else {
 		if ((mount_nfs = open_mount("nfs", MODPREFIX))) {
 			init_ctr++;
-			return 0;
 		} else {
 			kill_context(ctxt);
 			*context = NULL;
+			instance_mutex_unlock();
 			return 1;
 		}
-	} else {
-		init_ctr++;
-		return 0;
 	}
+	instance_mutex_unlock();
+	return 0;
 }
 
 static const char *parse_options(const char *str, char **ret, unsigned int logopt)
@@ -434,17 +444,23 @@ static char *concat_options(char *left, char *right)
 	char buf[MAX_ERR_BUF];
 	char *ret;
 
-	if (left == NULL || *left == '\0')
-		return strdup(right);
+	if (left == NULL || *left == '\0') {
+		ret = strdup(right);
+		free(right);
+		return ret;
+	}
 
-	if (right == NULL || *right == '\0')
-		return strdup(left);
+	if (right == NULL || *right == '\0') {
+		ret = strdup(left);
+		free(left);
+		return ret;
+	}
 
 	ret = malloc(strlen(left) + strlen(right) + 2);
 
 	if (ret == NULL) {
 		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-		error(LOGOPT_ANY, MODPREFIX "malloc: %s", estr);
+		logerr(MODPREFIX "malloc: %s", estr);
 		return NULL;
 	}
 
@@ -465,9 +481,11 @@ static int sun_mount(struct autofs_point *ap, const char *root,
 {
 	char *fstype = "nfs";	/* Default filesystem type */
 	int nonstrict = 1;
+	int use_weight_only = ap->flags & MOUNT_FLAG_USE_WEIGHT_ONLY;
 	int rv, cur_state;
 	char *mountpoint;
 	char *what;
+	char *type;
 
 	if (*options == '\0')
 		options = NULL;
@@ -510,6 +528,10 @@ static int sun_mount(struct autofs_point *ap, const char *root,
 					memcpy(np, cp, comma - cp + 1);
 					np += comma - cp + 1;
 				}
+			} else if (strncmp("no-use-weight-only", cp, 18) == 0) {
+				use_weight_only = -1;
+			} else if (strncmp("use-weight-only", cp, 15) == 0) {
+				use_weight_only = MOUNT_FLAG_USE_WEIGHT_ONLY;
 			} else if (strncmp("bg", cp, 2) == 0 ||
 				   strncmp("nofg", cp, 4) == 0) {
 				continue;
@@ -528,11 +550,10 @@ static int sun_mount(struct autofs_point *ap, const char *root,
 		options = noptions;
 	}
 
-
 	if (!strcmp(fstype, "autofs") && ctxt->macros) {
 		char *noptions = NULL;
 
-		if (!options) {
+		if (!options || *options == '\0') {
 			noptions = alloca(strlen(ctxt->macros) + 1);
 			*noptions = '\0';
 		} else {
@@ -545,7 +566,7 @@ static int sun_mount(struct autofs_point *ap, const char *root,
 			}
 		}
 
-		if (noptions) {
+		if (noptions && *noptions != '\0') {
 			strcat(noptions, ctxt->macros);
 			options = noptions;
 		} else {
@@ -557,12 +578,77 @@ static int sun_mount(struct autofs_point *ap, const char *root,
 	mountpoint = alloca(namelen + 1);
 	sprintf(mountpoint, "%.*s", namelen, name);
 
+	type = ap->entry->maps->type;
+	if (type && !strcmp(type, "hosts")) {
+		if (options && *options != '\0') {
+			int len = strlen(options);
+			int suid = strstr(options, "suid") ? 0 : 7;
+			int dev = strstr(options, "dev") ? 0 : 6;
+			int nointr = strstr(options, "nointr") ? 0 : 5;
+
+			if (suid || dev || nointr) {
+				char *tmp = alloca(len + suid + dev + nointr + 1);
+				if (!tmp) {
+					error(ap->logopt, MODPREFIX
+					      "alloca failed for options");
+					if (nonstrict)
+						return -1;
+					return 1;
+				}
+
+				strcpy(tmp, options);
+				if (suid)
+					strcat(tmp, ",nosuid");
+				if (dev)
+					strcat(tmp, ",nodev");
+				if (nointr)
+					strcat(tmp, ",intr");
+				options = tmp;
+			}
+		} else {
+			char *tmp = alloca(18);
+			if (!tmp) {
+				error(ap->logopt,
+				      MODPREFIX "alloca failed for options");
+				if (nonstrict)
+					return -1;
+				return 1;
+			}
+			strcpy(tmp, "nosuid,nodev,intr");
+			options = tmp;
+		}
+	}
+
 	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cur_state);
-	if (!strcmp(fstype, "nfs")) {
+	if (!strcmp(fstype, "nfs") || !strcmp(fstype, "nfs4")) {
 		what = alloca(loclen + 1);
 		memcpy(what, loc, loclen);
 		what[loclen] = '\0';
 
+		/* Add back "[no-]use-weight-only" for NFS mounts only */
+		if (use_weight_only) {
+			char *tmp;
+			int len;
+
+			if (options && *options != '\0') {
+				len = strlen(options) + 19;
+				tmp = alloca(len);
+				strcpy(tmp, options);
+				strcat(tmp, ",");
+				if (use_weight_only == MOUNT_FLAG_USE_WEIGHT_ONLY)
+					strcat(tmp, "use-weight-only");
+				else
+					strcat(tmp, "no-use-weight-only");
+			} else {
+				tmp = alloca(19);
+				if (use_weight_only == MOUNT_FLAG_USE_WEIGHT_ONLY)
+					strcpy(tmp, "use-weight-only");
+				else
+					strcpy(tmp, "no-use-weight-only");
+			}
+			options = tmp;
+		}
+
 		debug(ap->logopt, MODPREFIX
 		      "mounting root %s, mountpoint %s, "
 		      "what %s, fstype %s, options %s",
@@ -571,14 +657,18 @@ static int sun_mount(struct autofs_point *ap, const char *root,
 		rv = mount_nfs->mount_mount(ap, root, mountpoint, strlen(mountpoint),
 					    what, fstype, options, mount_nfs->context);
 	} else {
-		what = alloca(loclen + 1);
-		if (*loc == ':') {
-			loclen--;
-			memcpy(what, loc + 1, loclen);
-			what[loclen] = '\0';
-		} else {
-			memcpy(what, loc, loclen);
-			what[loclen] = '\0';
+		if (!loclen)
+			what = NULL;
+		else {
+			what = alloca(loclen + 1);
+			if (*loc == ':') {
+				loclen--;
+				memcpy(what, loc + 1, loclen);
+				what[loclen] = '\0';
+			} else {
+				memcpy(what, loc, loclen);
+				what[loclen] = '\0';
+			}
 		}
 
 		debug(ap->logopt, MODPREFIX
@@ -609,8 +699,7 @@ static int check_is_multi(const char *mapent)
 	int not_first_chunk = 0;
 
 	if (!p) {
-		crit(LOGOPT_ANY,
-		     MODPREFIX "unexpected NULL map entry pointer");
+		logerr(MODPREFIX "unexpected NULL map entry pointer");
 		return 0;
 	}
 
@@ -658,10 +747,10 @@ static int check_is_multi(const char *mapent)
 }
 
 static int
-add_offset_entry(struct autofs_point *ap, const char *name,
-		 const char *m_root, int m_root_len,
-		 const char *path, const char *myoptions, const char *loc,
-		 time_t age)
+update_offset_entry(struct autofs_point *ap, const char *name,
+		    const char *m_root, int m_root_len,
+		    const char *path, const char *myoptions, const char *loc,
+		    time_t age)
 {
 	struct map_source *source;
 	struct mapent_cache *mc;
@@ -676,7 +765,8 @@ add_offset_entry(struct autofs_point *ap, const char *name,
 
 	mc = source->mc;
 
-	if (!*path || !*loc) {
+	/* Internal hosts map may have loc == NULL */
+	if (!*path) {
 		error(ap->logopt,
 		      MODPREFIX "syntax error in offset %s -> %s", path, loc);
 		return CHE_FAIL;
@@ -684,8 +774,10 @@ add_offset_entry(struct autofs_point *ap, const char *name,
 
 	p_len = strlen(path);
 	/* Trailing '/' causes us pain */
-	if (p_len > 1 && path[p_len - 1] == '/')
-		p_len--;
+	if (p_len > 1) {
+		while (p_len > 1 && path[p_len - 1] == '/')
+			p_len--;
+	}
 	m_key_len = m_root_len + p_len;
 	if (m_key_len > PATH_MAX) {
 		error(ap->logopt, MODPREFIX "multi mount key too long");
@@ -708,23 +800,30 @@ add_offset_entry(struct autofs_point *ap, const char *name,
 	if (*myoptions) {
 		strcpy(m_mapent, "-");
 		strcat(m_mapent, myoptions);
-		strcat(m_mapent, " ");
-		strcat(m_mapent, loc);
+		if (loc) {
+			strcat(m_mapent, " ");
+			strcat(m_mapent, loc);
+		}
 	} else
 		strcpy(m_mapent, loc);
 
-	ret = cache_add_offset(mc, name, m_key, m_mapent, age);
-	if (ret == CHE_OK)
-		debug(ap->logopt, MODPREFIX
-		      "added multi-mount offset %s -> %s", path, m_mapent);
-	else
+	ret = cache_update_offset(mc, name, m_key, m_mapent, age);
+	if (ret == CHE_DUPLICATE)
 		warn(ap->logopt, MODPREFIX
-		      "syntax error or duplicate offset %s -> %s", path, loc);
+		     "syntax error or duplicate offset %s -> %s", path, loc);
+	else if (ret == CHE_FAIL)
+		debug(ap->logopt, MODPREFIX
+		      "failed to update multi-mount offset %s -> %s", path, m_mapent);
+	else {
+		ret = CHE_OK;
+		debug(ap->logopt, MODPREFIX
+		      "updated multi-mount offset %s -> %s", path, m_mapent);
+	}
 
 	return ret;
 }
 
-static int validate_location(char *loc)
+static int validate_location(unsigned int logopt, char *loc)
 {
 	char *ptr = loc;
 
@@ -732,23 +831,54 @@ static int validate_location(char *loc)
 	if (*ptr == ':')
 		return 1;
 
-	/* If a ':' is present now it must be a host name */
-	if (check_colon(ptr)) {
-		while (*ptr && *ptr != ':') {
+	/*
+	 * If a ':/' is present now it must be a host name, except
+	 * for those special file systems like sshfs which use "#"
+	 * and "@" in the host name part and ipv6 addresses that
+	 * have ":", "[" and "]".
+	 */
+	if (!check_colon(ptr)) {
+		char *esc;
+		/*
+		 * Don't forget cases where a colon is present but
+		 * not followed by a "/" or, if there is no colon at
+		 * all, we don't know if it is actually invalid since
+		 * it may be a map name by itself, for example.
+		 */
+		if (!strchr(ptr, ':') ||
+		    ((esc = strchr(ptr, '\\')) && *(esc + 1) == ':') ||
+		    !strncmp(ptr, "file:", 5) || !strncmp(ptr, "yp:", 3) ||
+		    !strncmp(ptr, "nis:", 4) || !strncmp(ptr, "nisplus:", 8) ||
+		    !strncmp(ptr, "ldap:", 5) || !strncmp(ptr, "ldaps:", 6) ||
+		    !strncmp(ptr, "sss:", 4) || !strncmp(ptr, "dir:", 4))
+			return 1;
+		error(logopt,
+		      "expected colon delimeter not found in location %s",
+		      loc);
+		return 0;
+	} else {
+		while (*ptr && strncmp(ptr, ":/", 2)) {
 			if (!(isalnum(*ptr) ||
 			    *ptr == '-' || *ptr == '.' || *ptr == '_' ||
-			    *ptr == ',' || *ptr == '(' || *ptr == ')'))
+			    *ptr == ',' || *ptr == '(' || *ptr == ')' ||
+			    *ptr == '#' || *ptr == '@' || *ptr == ':' ||
+			    *ptr == '[' || *ptr == ']')) {
+				error(logopt, "invalid character \"%c\" "
+				      "found in location %s", *ptr, loc);
 				return 0;
+			}
 			ptr++;
 		}
 
-		if (*ptr && *ptr == ':')
+		if (*ptr && !strncmp(ptr, ":/", 2))
 			ptr++;
 	}
 
 	/* Must always be something following */
-	if (!*ptr)
+	if (!*ptr) {
+		error(logopt, "invalid location %s", loc);
 		return 0;
+	}
 
 	return 1;
 }
@@ -775,18 +905,23 @@ static int parse_mapent(const char *ent, char *g_options, char **options, char *
 			char *tmp, *newopt = NULL;
 
 			p = parse_options(p, &newopt, logopt);
-			tmp = concat_options(myoptions, newopt);
-			if (!tmp) {
-				char *estr;
-				estr = strerror_r(errno, buf, MAX_ERR_BUF);
-				error(logopt, MODPREFIX
-				      "concat_options: %s", estr);
-				if (newopt)
-					free(newopt);
+			if (newopt && strstr(newopt, myoptions)) {
 				free(myoptions);
-				return 0;
+				myoptions = newopt;
+			} else {
+				tmp = concat_options(myoptions, newopt);
+				if (!tmp) {
+					char *estr;
+					estr = strerror_r(errno, buf, MAX_ERR_BUF);
+					error(logopt, MODPREFIX
+					      "concat_options: %s", estr);
+					if (newopt)
+						free(newopt);
+					free(myoptions);
+					return 0;
+				}
+				myoptions = tmp;
 			}
-			myoptions = tmp;
 
 			p = skipspace(p);
 		} while (*p == '-');
@@ -806,11 +941,11 @@ static int parse_mapent(const char *ent, char *g_options, char **options, char *
 	if (*p == '/') {
 		warn(logopt, MODPREFIX "error location begins with \"/\"");
 		free(myoptions);
+		free(loc);
 		return 0;
 	}
 
-	if (!validate_location(loc)) {
-		warn(logopt, MODPREFIX "invalid location");
+	if (!validate_location(logopt, loc)) {
 		free(myoptions);
 		free(loc);
 		return 0;
@@ -822,11 +957,11 @@ static int parse_mapent(const char *ent, char *g_options, char **options, char *
 	p = skipspace(p);
 
 	while (*p && ((*p == '"' && *(p + 1) != '/') || (*p != '"' && *p != '/'))) {
-		char *tmp, *ent;
+		char *tmp, *ent_chunk;
 
 		l = chunklen(p, check_colon(p));
-		ent = dequote(p, l, logopt);
-		if (!ent) {
+		ent_chunk = dequote(p, l, logopt);
+		if (!ent_chunk) {
 			warn(logopt, MODPREFIX "null location or out of memory");
 			free(myoptions);
 			free(loc);
@@ -837,26 +972,25 @@ static int parse_mapent(const char *ent, char *g_options, char **options, char *
 		if (*p == '/') {
 			warn(logopt,
 			      MODPREFIX "error location begins with \"/\"");
+			free(ent_chunk);
 			free(myoptions);
 			free(loc);
 			return 0;
 		}
 
-		if (!validate_location(ent)) {
-			warn(logopt,
-			      MODPREFIX "invalid location %s", ent);
-			free(ent);
+		if (!validate_location(logopt, ent_chunk)) {
+			free(ent_chunk);
 			free(myoptions);
 			free(loc);
 			return 0;
 		}
 
-		debug(logopt, MODPREFIX "dequote(\"%.*s\") -> %s", l, p, ent);
+		debug(logopt, MODPREFIX "dequote(\"%.*s\") -> %s", l, p, ent_chunk);
 
 		tmp = realloc(loc, strlen(loc) + l + 2);
 		if (!tmp) {
 			error(logopt, MODPREFIX "out of memory");
-			free(ent);
+			free(ent_chunk);
 			free(myoptions);
 			free(loc);
 			return 0;
@@ -864,9 +998,9 @@ static int parse_mapent(const char *ent, char *g_options, char **options, char *
 		loc = tmp;
 
 		strcat(loc, " ");
-		strcat(loc, ent);
+		strcat(loc, ent_chunk);
 
-		free(ent);
+		free(ent_chunk);
 
 		p += l;
 		p = skipspace(p);
@@ -878,53 +1012,174 @@ static int parse_mapent(const char *ent, char *g_options, char **options, char *
 	return (p - ent);
 }
 
-static int mount_subtree_offsets(struct autofs_point *ap, struct mapent_cache *mc, struct mapent *me)
+static void cleanup_multi_triggers(struct autofs_point *ap,
+			    struct mapent *me, const char *root, int start,
+			    const char *base)
+{
+	char path[PATH_MAX + 1];
+	char offset[PATH_MAX + 1];
+	char *poffset = offset;
+	struct mapent *oe;
+	struct list_head *mm_root, *pos;
+	const char o_root[] = "/";
+	const char *mm_base;
+
+	mm_root = &me->multi->multi_list;
+
+	if (!base)
+		mm_base = o_root;
+	else
+		mm_base = base;
+
+	pos = NULL;
+
+	/* Make sure "none" of the offsets have an active mount. */
+	while ((poffset = cache_get_offset(mm_base, poffset, start, mm_root, &pos))) {
+		oe = cache_lookup_offset(mm_base, poffset, start, &me->multi_list);
+		/* root offset is a special case */
+		if (!oe || !oe->mapent || (strlen(oe->key) - start) == 1)
+			continue;
+
+		strcpy(path, root);
+		strcat(path, poffset);
+		if (umount(path)) {
+			error(ap->logopt, "error recovering from mount fail");
+			error(ap->logopt, "cannot umount offset %s", path);
+		}
+	}
+
+	return;
+}
+
+static int mount_subtree(struct autofs_point *ap, struct mapent *me,
+			 const char *name, char *loc, char *options, void *ctxt)
 {
 	struct mapent *mm;
-	char *m_key;
-	int ret, start;
-	char *base, *m_root;
-	char buf[MAX_ERR_BUF];
+	struct mapent *ro;
+	char *mm_root, *mm_base, *mm_key;
+	const char *mnt_root;
+	unsigned int mm_root_len, mnt_root_len;
+	int start, ret = 0, rv;
+
+	rv = 0;
 
 	mm = me->multi;
+	mm_key = mm->key;
 
-	if (!mm)
-		return 0;
-
-	cache_multi_lock(me->parent);
-
-	m_key = mm->key;
-
-	if (*m_key == '/') {
-		m_root = m_key;
-		start = strlen(m_key);
+	if (*mm_key == '/') {
+		mm_root = mm_key;
+		start = strlen(mm_key);
 	} else {
-		start = strlen(ap->path) + strlen(m_key) + 1;
-		m_root = alloca(start + 1);
-		if (!m_root) {
-			char *estr;
-			cache_multi_unlock(me->parent);
-			estr = strerror_r(errno, buf, MAX_ERR_BUF);
-			error(ap->logopt, MODPREFIX "alloca: %s", estr);
-			return -1;
+		start = strlen(ap->path) + strlen(mm_key) + 1;
+		mm_root = alloca(start + 3);
+		strcpy(mm_root, ap->path);
+		strcat(mm_root, "/");
+		strcat(mm_root, mm_key);
+	}
+	mm_root_len = strlen(mm_root);
+
+	mnt_root = mm_root;
+	mnt_root_len = mm_root_len;
+
+	if (me == me->multi) {
+		/* name = NULL */
+		/* destination = mm_root */
+		mm_base = "/";
+
+		/* Mount root offset if it exists */
+		ro = cache_lookup_offset(mm_base, mm_base, strlen(mm_root), &me->multi_list);
+		if (ro) {
+			char *myoptions, *ro_loc, *tmp;
+			int namelen = name ? strlen(name) : 0;
+			const char *root;
+			int ro_len;
+
+			rv = parse_mapent(ro->mapent,
+				options, &myoptions, &ro_loc, ap->logopt);
+			if (!rv) {
+				warn(ap->logopt,
+				      MODPREFIX "failed to parse root offset");
+				cache_delete_offset_list(me->mc, name);
+				return 1;
+			}
+			ro_len = strlen(ro_loc);
+
+			tmp = alloca(mnt_root_len + 2);
+			strcpy(tmp, mnt_root);
+			tmp[mnt_root_len] = '/';
+			tmp[mnt_root_len + 1] = '\0';
+			root = tmp;
+
+			rv = sun_mount(ap, root, name, namelen, ro_loc, ro_len, myoptions, ctxt);
+
+			free(myoptions);
+			free(ro_loc);
+		}
+
+		if (ro && rv == 0) {
+			ret = mount_multi_triggers(ap, me, mnt_root, start, mm_base);
+			if (ret == -1) {
+				error(ap->logopt, MODPREFIX
+					 "failed to mount offset triggers");
+				cleanup_multi_triggers(ap, me, mnt_root, start, mm_base);
+				return 1;
+			}
+		} else if (rv <= 0) {
+			ret = mount_multi_triggers(ap, me, mm_root, start, mm_base);
+			if (ret == -1) {
+				error(ap->logopt, MODPREFIX
+					 "failed to mount offset triggers");
+				cleanup_multi_triggers(ap, me, mm_root, start, mm_base);
+				return 1;
+			}
+		}
+	} else {
+		int loclen = strlen(loc);
+		int namelen = strlen(name);
+
+		mnt_root = name;
+
+		/* name = mm_root + mm_base */
+		/* destination = mm_root + mm_base = name */
+		mm_base = &me->key[start];
+
+		rv = sun_mount(ap, mnt_root, name, namelen, loc, loclen, options, ctxt);
+		if (rv == 0) {
+			ret = mount_multi_triggers(ap, me->multi, mnt_root, start, mm_base);
+			if (ret == -1) {
+				error(ap->logopt, MODPREFIX
+					 "failed to mount offset triggers");
+				cleanup_multi_triggers(ap, me, mnt_root, start, mm_base);
+				return 1;
+			}
+		} else if (rv < 0) {
+			char *mm_root_base = alloca(strlen(mm_root) + strlen(mm_base) + 1);
+	
+			strcpy(mm_root_base, mm_root);
+			strcat(mm_root_base, mm_base);
+
+			ret = mount_multi_triggers(ap, me->multi, mm_root_base, start, mm_base);
+			if (ret == -1) {
+				error(ap->logopt, MODPREFIX
+					 "failed to mount offset triggers");
+				cleanup_multi_triggers(ap, me, mm_root, start, mm_base);
+				return 1;
+			}
 		}
-		strcpy(m_root, ap->path);
-		strcat(m_root, "/");
-		strcat(m_root, m_key);
 	}
 
-	base = &me->key[start];
+	/* Mount for base of tree failed */
+	if (rv > 0)
+		return rv;
 
-	ret = mount_multi_triggers(ap, m_root, me->multi, base);
-	if (ret == -1) {
-		cache_multi_unlock(me->parent);
-		error(ap->logopt, MODPREFIX "failed to mount offset triggers");
-		return -1;
-	}
+	/*
+	 * Convert fail on nonstrict, non-empty multi-mount
+	 * to success
+	 */
+	if (rv < 0 && ret > 0)
+		rv = 0;
 
-	cache_multi_unlock(me->parent);
-
-	return ret;
+	return rv;
 }
 
 /*
@@ -946,12 +1201,13 @@ int parse_mount(struct autofs_point *ap, const char *name,
 	char buf[MAX_ERR_BUF];
 	struct map_source *source;
 	struct mapent_cache *mc;
-	struct mapent *me, *ro;
+	struct mapent *me;
 	char *pmapent, *options;
 	const char *p;
 	int mapent_len, rv = 0;
-	int optlen, cur_state;
+	int cur_state;
 	int slashify = ctxt->slashify_colons;
+	unsigned int append_options;
 
 	source = ap->entry->current;
 	ap->entry->current = NULL;
@@ -981,7 +1237,7 @@ int parse_mount(struct autofs_point *ap, const char *name,
 	pmapent = alloca(mapent_len + 1);
 	if (!pmapent) {	
 		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-		error(ap->logopt, MODPREFIX "alloca: %s", estr);
+		logerr(MODPREFIX "alloca: %s", estr);
 		ctxt->subst = removestdenv(ctxt->subst);
 		macro_unlock();
 		pthread_setcancelstate(cur_state, NULL);
@@ -997,45 +1253,70 @@ int parse_mount(struct autofs_point *ap, const char *name,
 
 	debug(ap->logopt, MODPREFIX "expanded entry: %s", pmapent);
 
+	append_options = defaults_get_append_options();
 	options = strdup(ctxt->optstr ? ctxt->optstr : "");
 	if (!options) {
 		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-		error(ap->logopt, MODPREFIX "strdup: %s", estr);
+		logerr(MODPREFIX "strdup: %s", estr);
 		return 1;
 	}
-	optlen = strlen(options);
 
 	p = skipspace(pmapent);
 
 	/* Deal with 0 or more options */
 	if (*p == '-') {
-		char *mnt_options = NULL;
+		char *tmp, *mnt_options = NULL;
 
 		do {
-			char *tmp, *noptions = NULL;
+			char *noptions = NULL;
 
 			p = parse_options(p, &noptions, ap->logopt);
-			tmp = concat_options(mnt_options, noptions);
-			if (!tmp) {
-				char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-				error(ap->logopt,
-				      MODPREFIX "concat_options: %s", estr);
-				if (noptions)
-					free(noptions);
-				if (mnt_options)
-					free(mnt_options);
-				free(options);
-				return 1;
+			if (mnt_options && noptions && strstr(noptions, mnt_options)) {
+				free(mnt_options);
+				mnt_options = noptions;
+			} else {
+				tmp = concat_options(mnt_options, noptions);
+				if (!tmp) {
+					char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
+					error(ap->logopt,
+					      MODPREFIX "concat_options: %s", estr);
+					if (noptions)
+						free(noptions);
+					if (mnt_options)
+						free(mnt_options);
+					free(options);
+					return 1;
+				}
+				mnt_options = tmp;
 			}
-			mnt_options = tmp;
 
 			p = skipspace(p);
 		} while (*p == '-');
 
-		if (options)
+		if (options && !append_options) {
 			free(options);
+			options = NULL;
+		}
 
-		options = mnt_options;
+		if (append_options) {
+			if (options && mnt_options && strstr(mnt_options, options)) {
+				free(options);
+				options = mnt_options;
+			} else {
+				tmp = concat_options(options, mnt_options);
+				if (!tmp) {
+					char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
+					error(ap->logopt, MODPREFIX "concat_options: %s", estr);
+					if (options)
+						free(options);
+					if (mnt_options)
+						free(mnt_options);
+					return 1;
+				}
+				options = tmp;
+			}
+		} else
+			options = mnt_options;
 	}
 
 	debug(ap->logopt, MODPREFIX "gathered options: %s", options);
@@ -1043,8 +1324,8 @@ int parse_mount(struct autofs_point *ap, const char *name,
 	if (check_is_multi(p)) {
 		char *m_root = NULL;
 		int m_root_len;
-		time_t age = time(NULL);
-		int l, ret;
+		time_t age;
+		int l;
 
 		/* If name starts with "/" it's a direct mount */
 		if (*name == '/') {
@@ -1053,7 +1334,7 @@ int parse_mount(struct autofs_point *ap, const char *name,
 			if (!m_root) {
 				char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
 				free(options);
-				error(ap->logopt, MODPREFIX "alloca: %s", estr);
+				logerr(MODPREFIX "alloca: %s", estr);
 				return 1;
 			}
 			strcpy(m_root, name);
@@ -1063,7 +1344,7 @@ int parse_mount(struct autofs_point *ap, const char *name,
 			if (!m_root) {
 				char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
 				free(options);
-				error(ap->logopt, MODPREFIX "alloca: %s", estr);
+				logerr(MODPREFIX "alloca: %s", estr);
 				return 1;
 			}
 			strcpy(m_root, ap->path);
@@ -1071,34 +1352,6 @@ int parse_mount(struct autofs_point *ap, const char *name,
 			strcat(m_root, name);
 		}
 
-		cache_writelock(mc);
-		me = cache_lookup_distinct(mc, name);
-		if (!me) {
-			int ret;
-			/*
-			 * Not in the cache, perhaps it's a program map
-			 * or one that doesn't support enumeration
-			 */
-			ret = cache_add(mc, name, mapent, time(NULL));
-			if (ret == CHE_FAIL) {
-				cache_unlock(mc);
-				free(options);
-				return 1;
-			}
-		} else {
-			/*
-			 * If the entry exists it must not have any existing
-			 * multi-mount subordinate entries since we are
-			 * mounting this afresh. We need to do this to allow
-			 * us to fail on the check for duplicate offsets in
-			 * we don't know when submounts go away.
-			 */
-			cache_multi_lock(me);
-			cache_delete_offset_list(mc, name);
-			cache_multi_unlock(me);
-		}
-		cache_unlock(mc);
-
 		cache_readlock(mc);
 		me = cache_lookup_distinct(mc, name);
 		if (me) {
@@ -1115,8 +1368,10 @@ int parse_mount(struct autofs_point *ap, const char *name,
 			return 1;
 		}
 
+		age = me->age;
+
 		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cur_state);
-		cache_multi_lock(me);
+		cache_multi_writelock(me);
 		/* It's a multi-mount; deal with it */
 		do {
 			char *path, *myoptions, *loc;
@@ -1149,13 +1404,17 @@ int parse_mount(struct autofs_point *ap, const char *name,
 
 			l = parse_mapent(p, options, &myoptions, &loc, ap->logopt);
 			if (!l) {
-				cache_delete_offset_list(mc, name);
-				cache_multi_unlock(me);
-				cache_unlock(mc);
-				free(path);
-				free(options);
-				pthread_setcancelstate(cur_state, NULL);
-				return 1;
+				if (!(strstr(myoptions, "fstype=autofs") &&
+				      strstr(myoptions, "hosts"))) {
+					error(LOGOPT_ANY, "I think I'm a hosts map? l %d", l);
+					cache_delete_offset_list(mc, name);
+					cache_multi_unlock(me);
+					cache_unlock(mc);
+					free(path);
+					free(options);
+					pthread_setcancelstate(cur_state, NULL);
+					return 1;
+				}
 			}
 
 			p += l;
@@ -1164,9 +1423,9 @@ int parse_mount(struct autofs_point *ap, const char *name,
 			master_source_current_wait(ap->entry);
 			ap->entry->current = source;
 
-			status = add_offset_entry(ap, name,
-						m_root, m_root_len,
-						path, myoptions, loc, age);
+			status = update_offset_entry(ap, name,
+						     m_root, m_root_len,
+						     path, myoptions, loc, age);
 
 			if (status != CHE_OK) {
 				warn(ap->logopt, MODPREFIX "error adding multi-mount");
@@ -1188,52 +1447,14 @@ int parse_mount(struct autofs_point *ap, const char *name,
 
 		/*
 		 * We've got the ordered list of multi-mount entries so go
-		 * through and set the parent entry of each
+		 * through and remove any stale entries if this is the top
+		 * of the multi-mount and set the parent entry of each.
 		 */
+		if (me == me->multi)
+			clean_stale_multi_triggers(ap, me, NULL, NULL);
 		cache_set_parents(me);
 
-		/* Mount root offset if it exists */
-		ro = cache_lookup_offset("/", "/", strlen(m_root), &me->multi_list);
-		if (ro) {
-			char *myoptions, *loc;
-
-			rv = parse_mapent(ro->mapent,
-				options, &myoptions, &loc, ap->logopt);
-			if (!rv) {
-				warn(ap->logopt,
-				      MODPREFIX "failed to mount root offset");
-				cache_delete_offset_list(mc, name);
-				cache_multi_unlock(me);
-				cache_unlock(mc);
-				free(options);
-				pthread_setcancelstate(cur_state, NULL);
-				return 1;
-			}
-
-			rv = sun_mount(ap, m_root,
-				"/", 1, loc, strlen(loc), myoptions, ctxt);
-
-			free(myoptions);
-			free(loc);
-		}
-
-		ret = mount_multi_triggers(ap, m_root, me, "/");
-		if (ret == -1) {
-			warn(ap->logopt,
-			      MODPREFIX "failed to mount offset triggers");
-			cache_multi_unlock(me);
-			cache_unlock(mc);
-			free(options);
-			pthread_setcancelstate(cur_state, NULL);
-			return 1;
-		}
-
-		/*
-		 * Convert fail on nonstrict, non-empty multi-mount
-		 * to success
-		 */
-		if (rv < 0 && ret > 0)
-			rv = 0;
+		rv = mount_subtree(ap, me, name, NULL, options, ctxt);
 
 		cache_multi_unlock(me);
 		cache_unlock(mc);
@@ -1248,6 +1469,31 @@ int parse_mount(struct autofs_point *ap, const char *name,
 		int loclen;
 		int l;
 
+		/*
+		 * If this is an offset belonging to a multi-mount entry
+		 * it's already been parsed (above) and any option string
+		 * has already been stripped so just use the remainder.
+		 */
+		cache_readlock(mc);
+		if (*name == '/' &&
+		   (me = cache_lookup_distinct(mc, name)) && me->multi) {
+			loc = strdup(p);
+			if (!loc) {
+				free(options);
+				cache_unlock(mc);
+				warn(ap->logopt, MODPREFIX "out of memory");
+				return 1;
+			}
+			cache_multi_writelock(me);
+			rv = mount_subtree(ap, me, name, loc, options, ctxt);
+			cache_multi_unlock(me);
+			cache_unlock(mc);
+			free(loc);
+			free(options);
+			return rv;
+		}
+		cache_unlock(mc);
+
 		l = chunklen(p, check_colon(p));
 		loc = dequote(p, l, ap->logopt);
 		if (!loc) {
@@ -1259,15 +1505,15 @@ int parse_mount(struct autofs_point *ap, const char *name,
 		/* Location can't begin with a '/' */
 		if (*p == '/') {
 			free(options);
+			free(loc);
 			warn(ap->logopt,
 			      MODPREFIX "error location begins with \"/\"");
 			return 1;
 		}
 
-		if (!validate_location(loc)) {
+		if (!validate_location(ap->logopt, loc)) {
 			free(loc);
 			free(options);
-			warn(ap->logopt, MODPREFIX "invalid location");
 			return 1;
 		}
 
@@ -1290,11 +1536,10 @@ int parse_mount(struct autofs_point *ap, const char *name,
 				return 1;
 			}
 
-			if (!validate_location(ent)) {
+			if (!validate_location(ap->logopt, ent)) {
 				free(ent);
 				free(loc);
 				free(options);
-				warn(ap->logopt, MODPREFIX "invalid location");
 				return 1;
 			}
 
@@ -1320,38 +1565,38 @@ int parse_mount(struct autofs_point *ap, const char *name,
 			p = skipspace(p);
 		}
 
-		loclen = strlen(loc);
-		if (loclen == 0) {
-			free(loc);
-			free(options);
-			error(ap->logopt,
-			      MODPREFIX "entry %s is empty!", name);
-			return 1;
+		/*
+		 * If options are asking for a hosts map loc should be
+		 * NULL but we see it can contain junk, so ....
+		 */
+		if ((strstr(options, "fstype=autofs") &&
+		     strstr(options, "hosts"))) {
+			loc = NULL;
+			loclen = 0;
+		} else {
+			loclen = strlen(loc);
+			if (loclen == 0) {
+				free(loc);
+				free(options);
+				error(ap->logopt,
+				      MODPREFIX "entry %s is empty!", name);
+				return 1;
+			}
 		}
 
 		debug(ap->logopt,
 		      MODPREFIX "core of entry: options=%s, loc=%.*s",
 		      options, loclen, loc);
 
-		rv = sun_mount(ap, ap->path, name, name_len, loc, loclen, options, ctxt);
+		if (!strcmp(ap->path, "/-"))
+			rv = sun_mount(ap, name, name, name_len,
+				       loc, loclen, options, ctxt);
+		else
+			rv = sun_mount(ap, ap->path, name, name_len,
+				       loc, loclen, options, ctxt);
 
 		free(loc);
 		free(options);
-
-		/*
-		 * If it's a multi-mount insert the triggers
-		 * These are always direct mount triggers so root = ""
-		 */
-		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cur_state);
-		cache_readlock(mc);
-		me = cache_lookup_distinct(mc, name);
-		if (me) {
-			int ret = mount_subtree_offsets(ap, mc, me);
-			/* Convert fail on nonstrict, non-empty multi-mount to success */
-			if (rv < 0 && ret > 0)
-				rv = 0;
-		}
-		cache_unlock(mc);
 		pthread_setcancelstate(cur_state, NULL);
 	}
 	return rv;
@@ -1362,10 +1607,12 @@ int parse_done(void *context)
 	int rv = 0;
 	struct parse_context *ctxt = (struct parse_context *) context;
 
+	instance_mutex_lock();
 	if (--init_ctr == 0) {
 		rv = close_mount(mount_nfs);
 		mount_nfs = NULL;
 	}
+	instance_mutex_unlock();
 	if (ctxt)
 		kill_context(ctxt);
 
