@@ -52,6 +52,7 @@
 #include <net/if.h>
 #include <netinet/in.h>
 #include <netdb.h>
+#include <ifaddrs.h>
 
 #include "rpc_subs.h"
 #include "replicated.h"
@@ -110,58 +111,18 @@ void seed_random(void)
 	return;
 }
 
-static int alloc_ifreq(struct ifconf *ifc, int sock)
-{
-	int ret, lastlen = ifc_last_len, len = ifc_buf_len;
-	char err_buf[MAX_ERR_BUF], *buf;
-
-	while (1) {
-		buf = malloc(len);
-		if (!buf) {
-			char *estr = strerror_r(errno, err_buf, MAX_ERR_BUF);
-			logerr("malloc: %s", estr);
-			return 0;
-		}
-
-		ifc->ifc_len = len;
-		ifc->ifc_req = (struct ifreq *) buf;
-
-		ret = ioctl(sock, SIOCGIFCONF, ifc);
-		if (ret == -1) {
-			char *estr = strerror_r(errno, err_buf, MAX_ERR_BUF);
-			logerr("ioctl: %s", estr);
-			free(buf);
-			return 0;
-		}
-
-		if (ifc->ifc_len <= lastlen)
-			break;
-
-		lastlen = ifc->ifc_len;
-		len += MAX_IFC_BUF;
-		free(buf);
-	}
-
-	if (lastlen != ifc_last_len) {
-		ifc_last_len = lastlen;
-		ifc_buf_len = len;
-	}
-
-	return 1;
-}
-
 static unsigned int get_proximity(struct sockaddr *host_addr)
 {
+	struct ifaddrs *ifa = NULL;
+	struct ifaddrs *this;
 	struct sockaddr_in *addr, *msk_addr, *if_addr;
 	struct sockaddr_in6 *addr6, *msk6_addr, *if6_addr;
 	struct in_addr *hst_addr;
 	struct in6_addr *hst6_addr;
 	int addr_len;
-	char buf[MAX_ERR_BUF], *ptr;
-	struct ifconf ifc;
-	struct ifreq *ifr, nmptr;
-	int sock, ret, i;
+	char buf[MAX_ERR_BUF];
 	uint32_t mask, ha, ia, *mask6, *ha6, *ia6;
+	int ret;
 
 	addr = NULL;
 	addr6 = NULL;
@@ -170,23 +131,24 @@ static unsigned int get_proximity(struct sockaddr *host_addr)
 	mask6 = NULL;
 	ha6 = NULL;
 	ia6 = NULL;
+	ha = 0;
 
 	switch (host_addr->sa_family) {
 	case AF_INET:
 		addr = (struct sockaddr_in *) host_addr;
 		hst_addr = (struct in_addr *) &addr->sin_addr;
 		ha = ntohl((uint32_t) hst_addr->s_addr);
-		addr_len = sizeof(hst_addr);
+		addr_len = sizeof(*hst_addr);
 		break;
 
 	case AF_INET6:
-#ifndef INET6
+#ifndef WITH_LIBTIRPC
 		return PROXIMITY_UNSUPPORTED;
 #else
 		addr6 = (struct sockaddr_in6 *) host_addr;
 		hst6_addr = (struct in6_addr *) &addr6->sin6_addr;
 		ha6 = &hst6_addr->s6_addr32[0];
-		addr_len = sizeof(hst6_addr);
+		addr_len = sizeof(*hst6_addr);
 		break;
 #endif
 
@@ -194,97 +156,74 @@ static unsigned int get_proximity(struct sockaddr *host_addr)
 		return PROXIMITY_ERROR;
 	}
 
-	sock = open_sock(AF_INET, SOCK_DGRAM, 0);
-	if (sock < 0) {
+	ret = getifaddrs(&ifa);
+	if (ret) {
 		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-		logerr("socket creation failed: %s", estr);
+		logerr("getifaddrs: %s", estr);
 		return PROXIMITY_ERROR;
 	}
 
-	if (!alloc_ifreq(&ifc, sock)) {
-		close(sock);
-		return PROXIMITY_ERROR;
-	}
+	this = ifa;
+	while (this) {
+		if (!(this->ifa_flags & IFF_UP) ||
+		    this->ifa_flags & IFF_POINTOPOINT ||
+		    this->ifa_addr == NULL) {
+			this = this->ifa_next;
+			continue;
+		}
 
-	/* For each interface */
-
-	/* Is the address a local interface */
-	i = 0;
-	ptr = (char *) &ifc.ifc_buf[0];
-
-	while (ptr < (char *) ifc.ifc_req + ifc.ifc_len) {
-		ifr = (struct ifreq *) ptr;
-
-		switch (ifr->ifr_addr.sa_family) {
+		switch (this->ifa_addr->sa_family) {
 		case AF_INET:
-#ifndef INET6
 			if (host_addr->sa_family == AF_INET6)
 				break;
-#endif
-			if_addr = (struct sockaddr_in *) &ifr->ifr_addr;
+			if_addr = (struct sockaddr_in *) this->ifa_addr;
 			ret = memcmp(&if_addr->sin_addr, hst_addr, addr_len);
 			if (!ret) {
-				close(sock);
-				free(ifc.ifc_req);
+				freeifaddrs(ifa);
 				return PROXIMITY_LOCAL;
 			}
 			break;
 
 		case AF_INET6:
-#ifndef INET6
+#ifdef WITH_LIBTIRPC
 			if (host_addr->sa_family == AF_INET)
 				break;
-
-			if6_addr = (struct sockaddr_in6 *) &ifr->ifr_addr;
+			if6_addr = (struct sockaddr_in6 *) this->ifa_addr;
 			ret = memcmp(&if6_addr->sin6_addr, hst6_addr, addr_len);
 			if (!ret) {
-				close(sock);
-				free(ifc.ifc_req);
+				freeifaddrs(ifa);
 				return PROXIMITY_LOCAL;
 			}
 #endif
-
 		default:
 			break;
 		}
-
-		i++;
-		ptr = (char *) &ifc.ifc_req[i];
+		this = this->ifa_next;
 	}
 
-	i = 0;
-	ptr = (char *) &ifc.ifc_buf[0];
-
-	while (ptr < (char *) ifc.ifc_req + ifc.ifc_len) {
-		ifr = (struct ifreq *) ptr;
-
-		nmptr = *ifr;
-		ret = ioctl(sock, SIOCGIFNETMASK, &nmptr);
-		if (ret == -1) {
-			char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-			logerr("ioctl: %s", estr);
-			close(sock);
-			free(ifc.ifc_req);
-			return PROXIMITY_ERROR;
+	this = ifa;
+	while (this) {
+		if (!(this->ifa_flags & IFF_UP) ||
+		    this->ifa_flags & IFF_POINTOPOINT ||
+		    this->ifa_addr == NULL) {
+			this = this->ifa_next;
+			continue;
 		}
 
-		switch (ifr->ifr_addr.sa_family) {
+		switch (this->ifa_addr->sa_family) {
 		case AF_INET:
-#ifndef INET6
 			if (host_addr->sa_family == AF_INET6)
 				break;
-#endif
-			if_addr = (struct sockaddr_in *) &ifr->ifr_addr;
+			if_addr = (struct sockaddr_in *) this->ifa_addr;
 			ia =  ntohl((uint32_t) if_addr->sin_addr.s_addr);
 
-			/* Is the address within a localiy attached subnet */
+			/* Is the address within a localy attached subnet */
 
-			msk_addr = (struct sockaddr_in *) &nmptr.ifr_netmask;
+			msk_addr = (struct sockaddr_in *) this->ifa_netmask;
 			mask = ntohl((uint32_t) msk_addr->sin_addr.s_addr);
 
 			if ((ia & mask) == (ha & mask)) {
-				close(sock);
-				free(ifc.ifc_req);
+				freeifaddrs(ifa);
 				return PROXIMITY_SUBNET;
 			}
 
@@ -306,52 +245,45 @@ static unsigned int get_proximity(struct sockaddr *host_addr)
 				break;
 
 			if ((ia & mask) == (ha & mask)) {
-				close(sock);
-				free(ifc.ifc_req);
+				freeifaddrs(ifa);
 				return PROXIMITY_NET;
 			}
 			break;
 
 		case AF_INET6:
-#ifndef INET6
+#ifdef WITH_LIBTIRPC
 			if (host_addr->sa_family == AF_INET)
 				break;
-
-			if6_addr = (struct sockaddr_in6 *) &ifr->ifr_addr;
+			if6_addr = (struct sockaddr_in6 *) this->ifa_addr;
 			ia6 = &if6_addr->sin6_addr.s6_addr32[0];
 
 			/* Is the address within the network of the interface */
 
-			msk6_addr = (struct sockaddr_in6 *) &nmptr.ifr_netmask;
+			msk6_addr = (struct sockaddr_in6 *) this->ifa_netmask;
 			mask6 = &msk6_addr->sin6_addr.s6_addr32[0];
 
 			if (ipv6_mask_cmp(ha6, ia6, mask6)) {
-				close(sock);
-				free(ifc.ifc_req);
+				freeifaddrs(ifa);
 				return PROXIMITY_SUBNET;
 			}
 
 			/* How do we define "local network" in ipv6? */
 #endif
-			break;
-
 		default:
 			break;
 		}
-
-		i++;
-		ptr = (char *) &ifc.ifc_req[i];
+		this = this->ifa_next;
 	}
 
-	close(sock);
-	free(ifc.ifc_req);
+	freeifaddrs(ifa);
 
 	return PROXIMITY_OTHER;
 }
 
-static struct host *new_host(const char *name,
-			     struct sockaddr *addr, size_t addr_len,
-			     unsigned int proximity, unsigned int weight)
+struct host *new_host(const char *name,
+		      struct sockaddr *addr, size_t addr_len,
+		      unsigned int proximity, unsigned int weight,
+		      unsigned int options)
 {
 	struct host *new;
 	struct sockaddr *tmp2;
@@ -385,6 +317,7 @@ static struct host *new_host(const char *name,
 	new->addr = tmp2;
 	new->proximity = proximity;
 	new->weight = weight;
+	new->options = options;
 
 	return new;
 }
@@ -486,42 +419,14 @@ void free_host_list(struct host **list)
 	*list = NULL;
 }
 
-static unsigned short get_port_option(const char *options)
-{
-	const char *start;
-	long port = 0;
-
-	if (!options)
-		return NFS_PORT;
-
-	start = strstr(options, "port=");
-	if (!start)
-		port = NFS_PORT;
-	else {
-		char optport[30], *opteq, *end;
-		int len;
-
-		end = strchr(start, ',');
-		len = end ? end - start : strlen(start);
-		strncpy(optport, start, len);
-		optport[len] = '\0';
-		opteq = strchr(optport, '=');
-		if (opteq)
-			port = atoi(opteq + 1);
-	}
-
-	if (port < 0)
-		port = 0;
-
-	return (unsigned short) port;
-}
-
 static unsigned int get_nfs_info(unsigned logopt, struct host *host,
 			 struct conn_info *pm_info, struct conn_info *rpc_info,
-			 const char *proto, unsigned int version,
-			 const char *options, unsigned int random_selection)
+			 int proto, unsigned int version, int port)
 {
-	char *have_port_opt = options ? strstr(options, "port=") : NULL;
+	unsigned int random_selection = host->options & MOUNT_FLAG_RANDOM_SELECT;
+	unsigned int use_weight_only = host->options & MOUNT_FLAG_USE_WEIGHT_ONLY;
+	socklen_t len = INET6_ADDRSTRLEN;
+	char buf[len + 1];
 	struct pmap parms;
 	struct timeval start, end;
 	struct timezone tz;
@@ -529,34 +434,65 @@ static unsigned int get_nfs_info(unsigned logopt, struct host *host,
 	double taken = 0;
 	int status, count = 0;
 
-	debug(logopt,
-	      "called for host %s proto %s version 0x%x",
-	      host->name, proto, version);
+	if (host->addr)
+		debug(logopt, "called with host %s(%s) proto %d version 0x%x",
+		      host->name, get_addr_string(host->addr, buf, len),
+		      proto, version);
+	else
+		debug(logopt,
+		      "called for host %s proto %d version 0x%x",
+		      host->name, proto, version);
+
+	rpc_info->proto = proto;
+	if (port < 0)
+		rpc_info->port = NFS_PORT;
+	else if (port > 0)
+		rpc_info->port = port;
 
 	memset(&parms, 0, sizeof(struct pmap));
-
 	parms.pm_prog = NFS_PROGRAM;
-
-	/* Try to prode UDP first to conserve socket space */
-	rpc_info->proto = getprotobyname(proto);
-	if (!rpc_info->proto)
-		return 0;
+	parms.pm_prot = proto;
 
 	if (!(version & NFS4_REQUESTED))
 		goto v3_ver;
 
-	if (!(rpc_info->port = get_port_option(options)))
-		goto v3_ver;
+	if (!port) {
+		status = rpc_portmap_getclient(pm_info,
+				host->name, host->addr, host->addr_len,
+				proto, RPC_CLOSE_DEFAULT);
+		if (status == -EHOSTUNREACH) {
+			supported = status;
+			goto done_ver;
+		} else if (status)
+			goto done_ver;
+		parms.pm_vers = NFS4_VERSION;
+		status = rpc_portmap_getport(pm_info, &parms, &rpc_info->port);
+		if (status == -EHOSTUNREACH || status == -ETIMEDOUT) {
+			supported = status;
+			goto done_ver;
+		} else if (status < 0) {
+			if (version & NFS_VERS_MASK)
+				goto v3_ver; /* MOUNT_NFS_DEFAULT_PROTOCOL=4 */
+			else
+				goto done_ver;
+		}
+	}
 
-	if (rpc_info->proto->p_proto == IPPROTO_UDP)
+	if (rpc_info->proto == IPPROTO_UDP)
 		status = rpc_udp_getclient(rpc_info, NFS_PROGRAM, NFS4_VERSION);
 	else
 		status = rpc_tcp_getclient(rpc_info, NFS_PROGRAM, NFS4_VERSION);
-	if (status) {
+	if (status == -EHOSTUNREACH) {
+		supported = status;
+		goto done_ver;
+	} else if (!status) {
 		gettimeofday(&start, &tz);
 		status = rpc_ping_proto(rpc_info);
 		gettimeofday(&end, &tz);
-		if (status) {
+		if (status == -ETIMEDOUT) {
+			supported = status;
+			goto done_ver;
+		} else if (status > 0) {
 			double reply;
 			if (random_selection) {
 				/* Random value between 0 and 1 */
@@ -573,38 +509,49 @@ static unsigned int get_nfs_info(unsigned logopt, struct host *host,
 		}
 	}
 
-v3_ver:
-	if (!have_port_opt) {
-		status = rpc_portmap_getclient(pm_info,
-				host->name, host->addr, host->addr_len,
-				proto, RPC_CLOSE_DEFAULT);
-		if (!status)
-			goto done_ver;
-	}
+	if (!(version & NFS_VERS_MASK))
+		goto done_ver;
 
+v3_ver:
 	if (!(version & NFS3_REQUESTED))
 		goto v2_ver;
 
-	if (have_port_opt) {
-		if (!(rpc_info->port = get_port_option(options)))
+	if (!port && !pm_info->client) {
+		status = rpc_portmap_getclient(pm_info,
+				host->name, host->addr, host->addr_len,
+				proto, RPC_CLOSE_DEFAULT);
+		if (status == -EHOSTUNREACH) {
+			supported = status;
 			goto done_ver;
-	} else {
-		parms.pm_prot = rpc_info->proto->p_proto;
+		} else if (status)
+			goto done_ver;
+	}
+
+	if (!port) {
 		parms.pm_vers = NFS3_VERSION;
-		rpc_info->port = rpc_portmap_getport(pm_info, &parms);
-		if (!rpc_info->port)
+		status = rpc_portmap_getport(pm_info, &parms, &rpc_info->port);
+		if (status == -EHOSTUNREACH || status == -ETIMEDOUT) {
+			supported = status;
+			goto done_ver;
+		} else if (status < 0)
 			goto v2_ver;
 	}
 
-	if (rpc_info->proto->p_proto == IPPROTO_UDP)
+	if (rpc_info->proto == IPPROTO_UDP)
 		status = rpc_udp_getclient(rpc_info, NFS_PROGRAM, NFS3_VERSION);
 	else
 		status = rpc_tcp_getclient(rpc_info, NFS_PROGRAM, NFS3_VERSION);
-	if (status) {
+	if (status == -EHOSTUNREACH) {
+		supported = status;
+		goto done_ver;
+	} else if (!status) {
 		gettimeofday(&start, &tz);
 		status = rpc_ping_proto(rpc_info);
 		gettimeofday(&end, &tz);
-		if (status) {
+		if (status == -ETIMEDOUT) {
+			supported = status;
+			goto done_ver;
+		} else if (status > 0) {
 			double reply;
 			if (random_selection) {
 				/* Random value between 0 and 1 */
@@ -625,26 +572,41 @@ v2_ver:
 	if (!(version & NFS2_REQUESTED))
 		goto done_ver;
 
-	if (have_port_opt) {
-		if (!(rpc_info->port = get_port_option(options)))
+	if (!port && !pm_info->client) {
+		status = rpc_portmap_getclient(pm_info,
+				host->name, host->addr, host->addr_len,
+				proto, RPC_CLOSE_DEFAULT);
+		if (status == -EHOSTUNREACH) {
+			supported = status;
 			goto done_ver;
-	} else {
-		parms.pm_prot = rpc_info->proto->p_proto;
-		parms.pm_vers = NFS2_VERSION;
-		rpc_info->port = rpc_portmap_getport(pm_info, &parms);
-		if (!rpc_info->port)
+		} else if (status)
 			goto done_ver;
 	}
 
-	if (rpc_info->proto->p_proto == IPPROTO_UDP)
+	if (!port) {
+		parms.pm_vers = NFS2_VERSION;
+		status = rpc_portmap_getport(pm_info, &parms, &rpc_info->port);
+		if (status == -EHOSTUNREACH || status == -ETIMEDOUT) {
+			supported = status;
+			goto done_ver;
+		} else if (status < 0)
+			goto done_ver;
+	}
+
+	if (rpc_info->proto == IPPROTO_UDP)
 		status = rpc_udp_getclient(rpc_info, NFS_PROGRAM, NFS2_VERSION);
 	else
 		status = rpc_tcp_getclient(rpc_info, NFS_PROGRAM, NFS2_VERSION);
-	if (status) {
+	if (status == -EHOSTUNREACH) {
+		supported = status;
+		goto done_ver;
+	} else if (!status) {
 		gettimeofday(&start, &tz);
 		status = rpc_ping_proto(rpc_info);
 		gettimeofday(&end, &tz);
-		if (status) {
+		if (status == -ETIMEDOUT)
+			supported = status;
+		else if (status > 0) {
 			double reply;
 			if (random_selection) {
 				/* Random value between 0 and 1 */
@@ -662,7 +624,7 @@ v2_ver:
 	}
 
 done_ver:
-	if (rpc_info->proto->p_proto == IPPROTO_UDP) {
+	if (rpc_info->proto == IPPROTO_UDP) {
 		rpc_destroy_udp_client(rpc_info);
 		rpc_destroy_udp_client(pm_info);
 	} else {
@@ -675,7 +637,10 @@ done_ver:
 		 * Average response time to 7 significant places as
 		 * integral type.
 		 */
-		host->cost = (unsigned long) ((taken * 1000000) / count);
+		if (use_weight_only)
+			host->cost = 1;
+		else
+			host->cost = (unsigned long) ((taken * 1000000) / count);
 
 		/* Allow for user bias */
 		if (host->weight)
@@ -689,8 +654,7 @@ done_ver:
 }
 
 static int get_vers_and_cost(unsigned logopt, struct host *host,
-			     unsigned int version, const char *options,
-			     unsigned int random_selection)
+			     unsigned int version, int port)
 {
 	struct conn_info pm_info, rpc_info;
 	time_t timeout = RPC_TIMEOUT;
@@ -715,23 +679,28 @@ static int get_vers_and_cost(unsigned logopt, struct host *host,
 
 	vers &= version;
 
-	if (version & UDP_REQUESTED) {
+	if (version & TCP_REQUESTED) {
 		supported = get_nfs_info(logopt, host,
-					&pm_info, &rpc_info, "udp", vers,
-					options, random_selection);
-		if (supported) {
+				   &pm_info, &rpc_info, IPPROTO_TCP, vers, port);
+		if (IS_ERR(supported)) {
+			if (ERR(supported) == EHOSTUNREACH ||
+			    ERR(supported) == ETIMEDOUT)
+				return ret;
+		} else if (supported) {
 			ret = 1;
-			host->version |= (supported << 8);
+			host->version |= supported;
 		}
 	}
 
-	if (version & TCP_REQUESTED) {
+	if (version & UDP_REQUESTED) {
 		supported = get_nfs_info(logopt, host,
-					 &pm_info, &rpc_info, "tcp", vers,
-					 options, random_selection);
-		if (supported) {
+				   &pm_info, &rpc_info, IPPROTO_UDP, vers, port);
+		if (IS_ERR(supported)) {
+			if (!ret && ERR(supported) == ETIMEDOUT)
+				return ret;
+		} else if (supported) {
 			ret = 1;
-			host->version |= supported;
+			host->version |= (supported << 8);
 		}
 	}
 
@@ -739,26 +708,31 @@ static int get_vers_and_cost(unsigned logopt, struct host *host,
 }
 
 static int get_supported_ver_and_cost(unsigned logopt, struct host *host,
-				      unsigned int version, const char *options,
-				      unsigned int random_selection)
+				      unsigned int version, int port)
 {
-	char *have_port_opt = options ? strstr(options, "port=") : NULL;
+	unsigned int random_selection = host->options & MOUNT_FLAG_RANDOM_SELECT;
+	unsigned int use_weight_only = host->options & MOUNT_FLAG_USE_WEIGHT_ONLY;
+	socklen_t len = INET6_ADDRSTRLEN;
+	char buf[len + 1];
 	struct conn_info pm_info, rpc_info;
-	struct pmap parms;
-	const char *proto;
+	int proto;
 	unsigned int vers;
 	struct timeval start, end;
 	struct timezone tz;
 	double taken = 0;
 	time_t timeout = RPC_TIMEOUT;
-	int status;
+	int status = 0;
 
-	debug(logopt,
-	      "called with host %s version 0x%x", host->name, version);
+	if (host->addr)
+		debug(logopt, "called with host %s(%s) version 0x%x",
+			host->name, get_addr_string(host->addr, buf, len),
+			version);
+	else
+		debug(logopt, "called with host %s version 0x%x",
+			host->name, version);
 
 	memset(&pm_info, 0, sizeof(struct conn_info));
 	memset(&rpc_info, 0, sizeof(struct conn_info));
-	memset(&parms, 0, sizeof(struct pmap));
 
 	if (host->proximity == PROXIMITY_NET)
 		timeout = RPC_TIMEOUT * 2;
@@ -773,8 +747,6 @@ static int get_supported_ver_and_cost(unsigned logopt, struct host *host,
 	rpc_info.close_option = RPC_CLOSE_DEFAULT;
 	rpc_info.client = NULL;
 
-	parms.pm_prog = NFS_PROGRAM;
-
 	/*
 	 *  The version passed in is the version as defined in
 	 *  include/replicated.h.  However, the version we want to send
@@ -782,10 +754,10 @@ static int get_supported_ver_and_cost(unsigned logopt, struct host *host,
 	 *  So, we do the conversion here.
 	 */
 	if (version & UDP_SELECTED_MASK) {
-		proto = "udp";
+		proto = IPPROTO_UDP;
 		version >>= 8;
 	} else
-		proto = "tcp";
+		proto = IPPROTO_TCP;
 
 	switch (version) {
 	case NFS2_SUPPORTED:
@@ -802,38 +774,40 @@ static int get_supported_ver_and_cost(unsigned logopt, struct host *host,
 		return 0;
 	}
 
-	rpc_info.proto = getprotobyname(proto);
-	if (!rpc_info.proto)
-		return 0;
+	rpc_info.proto = proto;
 
-	status = 0;
-
-	parms.pm_vers = vers;
-	if (have_port_opt || (vers & NFS4_VERSION)) {
-		if (!(rpc_info.port = get_port_option(options)))
-			return 0;
-	} else {
+	if (port > 0)
+		rpc_info.port = port;
+	else if (vers & NFS4_VERSION && port < 0)
+		rpc_info.port = NFS_PORT;
+	else {
+		struct pmap parms;
 		int ret = rpc_portmap_getclient(&pm_info,
 				host->name, host->addr, host->addr_len,
 				proto, RPC_CLOSE_DEFAULT);
-		if (!ret)
+		if (ret)
 			return 0;
 
-		parms.pm_prot = rpc_info.proto->p_proto;
-		rpc_info.port = rpc_portmap_getport(&pm_info, &parms);
-		if (!rpc_info.port)
+		memset(&parms, 0, sizeof(struct pmap));
+		parms.pm_prog = NFS_PROGRAM;
+		parms.pm_prot = rpc_info.proto;
+		parms.pm_vers = vers;
+		ret = rpc_portmap_getport(&pm_info, &parms, &rpc_info.port);
+		if (ret < 0)
 			goto done;
 	}
 
-	if (rpc_info.proto->p_proto == IPPROTO_UDP)
-		status = rpc_udp_getclient(&rpc_info, NFS_PROGRAM, parms.pm_vers);
+	if (rpc_info.proto == IPPROTO_UDP)
+		status = rpc_udp_getclient(&rpc_info, NFS_PROGRAM, vers);
 	else
-		status = rpc_tcp_getclient(&rpc_info, NFS_PROGRAM, parms.pm_vers);
-	if (status) {
+		status = rpc_tcp_getclient(&rpc_info, NFS_PROGRAM, vers);
+	if (status == -EHOSTUNREACH)
+		goto done;
+	else if (!status) {
 		gettimeofday(&start, &tz);
 		status = rpc_ping_proto(&rpc_info);
 		gettimeofday(&end, &tz);
-		if (status) {
+		if (status > 0) {
 			if (random_selection) {
 				/* Random value between 0 and 1 */
 				taken = ((float) random())/((float) RAND_MAX+1);
@@ -845,7 +819,7 @@ static int get_supported_ver_and_cost(unsigned logopt, struct host *host,
 		}
 	}
 done:
-	if (rpc_info.proto->p_proto == IPPROTO_UDP) {
+	if (rpc_info.proto == IPPROTO_UDP) {
 		rpc_destroy_udp_client(&rpc_info);
 		rpc_destroy_udp_client(&pm_info);
 	} else {
@@ -855,7 +829,10 @@ done:
 
 	if (status) {
 		/* Response time to 7 significant places as integral type. */
-		host->cost = (unsigned long) (taken * 1000000);
+		if (use_weight_only)
+			host->cost = 1;
+		else
+			host->cost = (unsigned long) (taken * 1000000);
 
 		/* Allow for user bias */
 		if (host->weight)
@@ -870,8 +847,7 @@ done:
 }
 
 int prune_host_list(unsigned logopt, struct host **list,
-		    unsigned int vers, const char *options,
-		    unsigned int random_selection)
+		    unsigned int vers, int port)
 {
 	struct host *this, *last, *first;
 	struct host *new = NULL;
@@ -880,6 +856,7 @@ int prune_host_list(unsigned logopt, struct host **list,
 	unsigned int v2_udp_count, v3_udp_count, v4_udp_count;
 	unsigned int max_udp_count, max_tcp_count, max_count;
 	int status;
+	int kern_vers;
 
 	if (!*list)
 		return 0;
@@ -892,19 +869,34 @@ int prune_host_list(unsigned logopt, struct host **list,
 	this = first;
 	while (this && this->proximity == PROXIMITY_LOCAL)
 		this = this->next;
+	first = this;
 
 	/*
 	 * Check for either a list containing only proximity local hosts
 	 * or a single host entry whose proximity isn't local. If so
 	 * return immediately as we don't want to add probe latency for
 	 * the common case of a single filesystem mount request.
+	 *
+	 * But, if the kernel understands text nfs mount options then
+	 * mount.nfs most likely bypasses its probing and lets the kernel
+	 * do all the work. This can lead to long timeouts for hosts that
+	 * are not available so check the kernel version and mount.nfs
+	 * version and probe singleton mounts if the kernel version is
+	 * greater than 2.6.22 and mount.nfs version is greater than 1.1.1.
+	 * But also allow the MOUNT_WAIT configuration parameter to override
+	 * the probing.
 	 */
-	if (!this || !this->next)
-		return 1;
+	if (nfs_mount_uses_string_options &&
+	    defaults_get_mount_wait() == -1 &&
+	   (kern_vers = linux_version_code()) > KERNEL_VERSION(2, 6, 22)) {
+		if (!this)
+			return 1;
+	} else {
+		if (!this || !this->next)
+			return 1;
+	}
 
 	proximity = this->proximity;
-	first = this;
-	this = first;
 	while (this) {
 		struct host *next = this->next;
 
@@ -912,8 +904,7 @@ int prune_host_list(unsigned logopt, struct host **list,
 			break;
 
 		if (this->name) {
-			status = get_vers_and_cost(logopt, this, vers,
-						   options, random_selection);
+			status = get_vers_and_cost(logopt, this, vers, port);
 			if (!status) {
 				if (this == first) {
 					first = next;
@@ -1022,8 +1013,7 @@ int prune_host_list(unsigned logopt, struct host **list,
 			add_host(&new, this);
 		} else {
 			status = get_supported_ver_and_cost(logopt, this,
-						selected_version, options,
-						random_selection);
+						selected_version, port);
 			if (status) {
 				this->version = selected_version;
 				remove_host(list, this);
@@ -1041,13 +1031,27 @@ int prune_host_list(unsigned logopt, struct host **list,
 
 static int add_new_host(struct host **list,
 			const char *host, unsigned int weight,
-			struct addrinfo *host_addr)
+			struct addrinfo *host_addr,
+			unsigned int rr, unsigned int options)
 {
 	struct host *new;
 	unsigned int prx;
 	int addr_len;
 
 	prx = get_proximity(host_addr->ai_addr);
+
+	/*
+	 * If we want the weight to be the determining factor
+	 * when selecting a host, or we are using random selection,
+	 * then all hosts must have the same proximity. However,
+	 * if this is the local machine it should always be used
+	 * since it is certainly available.
+	 */
+	if (prx != PROXIMITY_LOCAL &&
+	   (options & (MOUNT_FLAG_USE_WEIGHT_ONLY |
+		       MOUNT_FLAG_RANDOM_SELECT)))
+		prx = PROXIMITY_SUBNET;
+
 	/*
 	 * If we tried to add an IPv6 address and we don't have IPv6
 	 * support return success in the hope of getting an IPv4
@@ -1058,8 +1062,14 @@ static int add_new_host(struct host **list,
 	if (prx == PROXIMITY_ERROR)
 		return 0;
 
-	addr_len = sizeof(struct sockaddr);
-	new = new_host(host, host_addr->ai_addr, addr_len, prx, weight);
+	if (host_addr->ai_addr->sa_family == AF_INET)
+		addr_len = INET_ADDRSTRLEN;
+	else if (host_addr->ai_addr->sa_family == AF_INET6)
+		addr_len = INET6_ADDRSTRLEN;
+	else
+		return 0;
+
+	new = new_host(host, host_addr->ai_addr, addr_len, prx, weight, options);
 	if (!new)
 		return 0;
 
@@ -1067,27 +1077,47 @@ static int add_new_host(struct host **list,
 		free_host(new);
 		return 0;
 	}
+	new->rr = rr;
 
 	return 1;
 }
 
-static int add_host_addrs(struct host **list, const char *host, unsigned int weight)
+static int add_host_addrs(struct host **list, const char *host,
+			  unsigned int weight, unsigned int options)
 {
 	struct addrinfo hints, *ni, *this;
+	char *n_ptr;
+	char *name = n_ptr = strdup(host);
+	int len;
+	char buf[MAX_ERR_BUF];
+	int rr = 0, rr4 = 0, rr6 = 0;
 	int ret;
 
+	if (!name) {
+		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
+		error(LOGOPT_ANY, "strdup: %s", estr);
+		error(LOGOPT_ANY, "failed to add host %s", host);
+		return 0;
+	}
+	len = strlen(name);
+
+	if (name[0] == '[' && name[--len] == ']') {
+		name[len] = '\0';
+		name++;
+	}
+
 	memset(&hints, 0, sizeof(hints));
 	hints.ai_flags = AI_NUMERICHOST;
 	hints.ai_family = AF_UNSPEC;
 	hints.ai_socktype = SOCK_DGRAM;
 
-	ret = getaddrinfo(host, NULL, &hints, &ni);
+	ret = getaddrinfo(name, NULL, &hints, &ni);
 	if (ret)
 		goto try_name;
 
 	this = ni;
 	while (this) {
-		ret = add_new_host(list, host, weight, this);
+		ret = add_new_host(list, host, weight, this, 0, options);
 		if (!ret)
 			break;
 		this = this->ai_next;
@@ -1101,22 +1131,39 @@ try_name:
 	hints.ai_family = AF_UNSPEC;
 	hints.ai_socktype = SOCK_DGRAM;
 
-	ret = getaddrinfo(host, NULL, &hints, &ni);
+	ret = getaddrinfo(name, NULL, &hints, &ni);
 	if (ret) {
 		error(LOGOPT_ANY, "hostname lookup failed: %s",
 		      gai_strerror(ret));
+		free(name);
 		return 0;
 	}
 
+	this = ni;
+	while (this->ai_next) {
+		if (this->ai_family == AF_INET) {
+			struct sockaddr_in *addr = (struct sockaddr_in *) this->ai_addr;
+			if (addr->sin_addr.s_addr != INADDR_LOOPBACK)
+				rr4++;
+		} else if (this->ai_family == AF_INET6) {
+			struct sockaddr_in6 *addr = (struct sockaddr_in6 *) this->ai_addr;
+			if (!IN6_IS_ADDR_LOOPBACK(addr->sin6_addr.s6_addr32))
+				rr6++;
+		}
+		this = this->ai_next;
+	}
+	if (rr4 > 1 || rr6 > 1)
+		rr++;
 	this = ni;
 	while (this) {
-		ret = add_new_host(list, host, weight, this);
+		ret = add_new_host(list, host, weight, this, rr, options);
 		if (!ret)
 			break;
 		this = this->ai_next;
 	}
 	freeaddrinfo(ni);
 done:
+	free(n_ptr);
 	return ret;
 }
 
@@ -1181,7 +1228,7 @@ static char *seek_delim(const char *s)
 	char *delim;
 
 	delim = strpbrk(p, "(, \t:");
-	if (delim && *delim != ':')
+	if (delim && *delim != ':' && (delim == s || *(delim - 1) != '\\'))
 		return delim;
 
 	while (*p) {
@@ -1197,7 +1244,8 @@ static char *seek_delim(const char *s)
 	return NULL;
 }
 
-int parse_location(unsigned logopt, struct host **hosts, const char *list)
+int parse_location(unsigned logopt, struct host **hosts,
+		   const char *list, unsigned int options)
 {
 	char *str, *p, *delim;
 	unsigned int empty = 1;
@@ -1229,6 +1277,12 @@ int parse_location(unsigned logopt, struct host **hosts, const char *list)
 					*delim = '\0';
 					weight = atoi(w);
 				}
+				else {
+					/* syntax error - Mismatched brackets */
+					free_host_list(hosts);
+					free(str);
+					return 0;
+				}
 				delim++;
 			}
 
@@ -1252,7 +1306,7 @@ int parse_location(unsigned logopt, struct host **hosts, const char *list)
 				}
 
 				if (p != delim) {
-					if (!add_host_addrs(hosts, p, weight)) {
+					if (!add_host_addrs(hosts, p, weight, options)) {
 						if (empty) {
 							p = next;
 							continue;
@@ -1274,7 +1328,7 @@ int parse_location(unsigned logopt, struct host **hosts, const char *list)
 				*delim = '\0';
 				next = delim + 1;
 
-				if (!add_host_addrs(hosts, p, weight)) {
+				if (!add_host_addrs(hosts, p, weight, options)) {
 					p = next;
 					continue;
 				}
